'''
    Added a simple camera class to an existing example.
    The camera class is built using following tutorials:
       https://learnopengl.com/Getting-started/Camera
       http://in2gpu.com/2016/03/14/opengl-fps-camera-quaternion/

    Controls:
        Move:
            Forward - W
            Backwards - S

        Strafe:
            Up - up arrow
            Down - down arrow
            Left - A
            Right - D

        Rotate:
            Left - Q
            Right - E

        Zoom:
            In - X
            Out - Z

    adopted by: Alex Zakrividoroga
'''

import numpy as np
from pyrr import Matrix44, Quaternion, Vector3, vector

import moderngl
from ported._example import Example


class Camera():

    def __init__(self, ratio):
        self._zoom_step = 0.1
        self._move_vertically = 0.1
        self._move_horizontally = 0.1
        self._rotate_horizontally = 0.1
        self._rotate_vertically = 0.1

        self._field_of_view_degrees = 60.0
        self._z_near = 0.1
        self._z_far = 100
        self._ratio = ratio
        self.build_projection()

        self._camera_position = Vector3([0.0, 0.0, -40.0])
        self._camera_front = Vector3([0.0, 0.0, 1.0])
        self._camera_up = Vector3([0.0, 1.0, 0.0])
        self._cameras_target = (self._camera_position + self._camera_front)
        self.build_look_at()

    def zoom_in(self):
        self._field_of_view_degrees = self._field_of_view_degrees - self._zoom_step
        self.build_projection()

    def zoom_out(self):
        self._field_of_view_degrees = self._field_of_view_degrees + self._zoom_step
        self.build_projection()

    def move_forward(self):
        self._camera_position = self._camera_position + self._camera_front * self._move_horizontally
        self.build_look_at()

    def move_backwards(self):
        self._camera_position = self._camera_position - self._camera_front * self._move_horizontally
        self.build_look_at()

    def strafe_left(self):
        self._camera_position = self._camera_position - vector.normalize(self._camera_front ^ self._camera_up) * self._move_horizontally
        self.build_look_at()

    def strafe_right(self):
        self._camera_position = self._camera_position + vector.normalize(self._camera_front ^ self._camera_up) * self._move_horizontally
        self.build_look_at()

    def strafe_up(self):
        self._camera_position = self._camera_position + self._camera_up * self._move_vertically
        self.build_look_at()

    def strafe_down(self):
        self._camera_position = self._camera_position - self._camera_up * self._move_vertically
        self.build_look_at()

    def rotate_left(self):
        rotation = Quaternion.from_y_rotation(2 * float(self._rotate_horizontally) * np.pi / 180)
        self._camera_front = rotation * self._camera_front
        self.build_look_at()

    def rotate_right(self):
        rotation = Quaternion.from_y_rotation(-2 * float(self._rotate_horizontally) * np.pi / 180)
        self._camera_front = rotation * self._camera_front
        self.build_look_at()

    def build_look_at(self):
        self._cameras_target = (self._camera_position + self._camera_front)
        self.mat_lookat = Matrix44.look_at(
            self._camera_position,
            self._cameras_target,
            self._camera_up)

    def build_projection(self):
        self.mat_projection = Matrix44.perspective_projection(
            self._field_of_view_degrees,
            self._ratio,
            self._z_near,
            self._z_far)


def grid(size, steps):
    u = np.repeat(np.linspace(-size, size, steps), 2)
    v = np.tile([-size, size], steps)
    w = np.zeros(steps * 2)
    return np.concatenate([np.dstack([u, v, w]), np.dstack([v, u, w])])


class PerspectiveProjection(Example):
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_vert;

                void main() {
                    gl_Position = Mvp * vec4(in_vert, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.1, 0.1, 0.1, 1.0);
                }
            ''',
        )

        self.camera = Camera(self.aspect_ratio)
        self.mvp = self.prog['Mvp']
        self.vbo = self.ctx.buffer(grid(15, 10).astype('f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

        self.states = {
            self.wnd.keys.W: False,     # forward
            self.wnd.keys.S: False,     # backwards
            self.wnd.keys.UP: False,    # strafe Up
            self.wnd.keys.DOWN: False,  # strafe Down
            self.wnd.keys.A: False,     # strafe left
            self.wnd.keys.D: False,     # strafe right
            self.wnd.keys.Q: False,     # rotate left
            self.wnd.keys.E: False,     # rotare right
            self.wnd.keys.Z: False,     # zoom in
            self.wnd.keys.X: False,     # zoom out
        }

    def move_camera(self):
        if self.states.get(self.wnd.keys.W):
            self.camera.move_forward()

        if self.states.get(self.wnd.keys.S):
            self.camera.move_backwards()

        if self.states.get(self.wnd.keys.UP):
            self.camera.strafe_up()

        if self.states.get(self.wnd.keys.DOWN):
            self.camera.strafe_down()

        if self.states.get(self.wnd.keys.A):
            self.camera.strafe_left()

        if self.states.get(self.wnd.keys.D):
            self.camera.strafe_right()

        if self.states.get(self.wnd.keys.Q):
            self.camera.rotate_left()

        if self.states.get(self.wnd.keys.E):
            self.camera.rotate_right()

        if self.states.get(self.wnd.keys.Z):
            self.camera.zoom_in()

        if self.states.get(self.wnd.keys.X):
            self.camera.zoom_out()

    def key_event(self, key, action, modifiers):
        if key not in self.states:
            print(key, action)
            return

        if action == self.wnd.keys.ACTION_PRESS:
            self.states[key] = True
        else:
            self.states[key] = False

    def render(self, time, frame_time):
        self.move_camera()

        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        self.mvp.write((self.camera.mat_projection * self.camera.mat_lookat).astype('f4'))
        self.vao.render(moderngl.LINES)


if __name__ == '__main__':
    PerspectiveProjection.run()



#!/usr/bin/env python3
'''Simple example of using tessellation to render a cubic Bézier curve'''

import numpy as np

import moderngl
from ported._example import Example


class Tessellation(Example):
    title = "Tessellation"
    gl_version = (4, 0)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
            #version 400 core

            in vec2 in_pos;

            void main() { gl_Position = vec4(in_pos, 0.0, 1.0); }
            ''',
            tess_control_shader='''
            #version 400 core

            layout(vertices = 4) out;

            void main() {
            // set tessellation levels, TODO compute dynamically
            gl_TessLevelOuter[0] = 1;
            gl_TessLevelOuter[1] = 32;

            // pass through vertex positions
            gl_out[gl_InvocationID].gl_Position = gl_in[gl_InvocationID].gl_Position;
            }
            ''',
            tess_evaluation_shader='''
            #version 400 core

            layout(isolines, fractional_even_spacing, ccw) in;

            // compute a point on a bezier curve with the points p0, p1, p2, p3
            // the parameter u is in [0, 1] and determines the position on the curve
            vec3 bezier(float u, vec3 p0, vec3 p1, vec3 p2, vec3 p3) {
            float B0 = (1.0 - u) * (1.0 - u) * (1.0 - u);
            float B1 = 3.0 * (1.0 - u) * (1.0 - u) * u;
            float B2 = 3.0 * (1.0 - u) * u * u;
            float B3 = u * u * u;

            return B0 * p0 + B1 * p1 + B2 * p2 + B3 * p3;
            }

            void main() {
            float u = gl_TessCoord.x;

            vec3 p0 = vec3(gl_in[0].gl_Position);
            vec3 p1 = vec3(gl_in[1].gl_Position);
            vec3 p2 = vec3(gl_in[2].gl_Position);
            vec3 p3 = vec3(gl_in[3].gl_Position);

            gl_Position = vec4(bezier(u, p0, p1, p2, p3), 1.0);
            }
            ''',
            fragment_shader='''
            #version 400 core

            out vec4 frag_color;

            void main() { frag_color = vec4(1.0); }
            '''
        )

        # four vertices define a cubic Bézier curve; has to match the shaders
        self.ctx.patch_vertices = 4

        self.ctx.line_width = 5.0
        vertices = np.array([
            [-1.0, 0.0],
            [-0.5, 1.0],
            [0.5, -1.0],
            [1.0, 0.0],
        ])

        vbo = self.ctx.buffer(vertices.astype('f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, vbo, 'in_pos')

    def render(self, time, frame_time):
        self.ctx.clear(0.2, 0.4, 0.7)
        self.vao.render(mode=moderngl.PATCHES)


if __name__ == '__main__':
    Tessellation.run()



from array import array
import struct
import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context(require=430)
        if not cls.ctx:
            raise unittest.SkipTest('compute shaders not supported')

    def test_1(self):
        compute_shader = self.ctx.compute_shader('''
            #version 430

            layout (local_size_x = 1, local_size_y = 1) in;

            layout (std430, binding = 1) buffer Input {
                float v1[4];
            };

            layout (std430, binding = 2) buffer Output {
                float v2[4];
            };

            void main() {
                v2[0] = v1[3];
                v2[1] = v1[2];
                v2[2] = v1[1];
                v2[3] = v1[0];
            }
        ''')

        buf1 = self.ctx.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))
        buf2 = self.ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))

        buf1.bind_to_storage_buffer(1)
        buf2.bind_to_storage_buffer(2)

        compute_shader.run()

        a, b, c, d = struct.unpack('4f', buf2.read())

        self.assertAlmostEqual(a, 4.0)
        self.assertAlmostEqual(b, 3.0)
        self.assertAlmostEqual(c, 2.0)
        self.assertAlmostEqual(d, 1.0)

    def test_image(self):
        texture = self.ctx.texture((100, 100), 4)
        texture.bind_to_image(0, read=True, write=True)
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')
        texture.release()

    def test_image_float(self):
        texture = self.ctx.texture((100, 100), 4, dtype='f4')
        texture.bind_to_image(0, read=True, write=True)
        self.assertEqual(self.ctx.error, 'GL_NO_ERROR')
        texture.release()

    def test_image_wrong_format(self):
        texture = self.ctx.texture((100, 100), 4)
        texture.bind_to_image(0, read=True, write=True, format=13371337)
        self.assertEqual(self.ctx.error, 'GL_INVALID_VALUE')
        texture.release()

    def test_3d_image(self):
        program = self.program = self.ctx.compute_shader(
            """
            #version 430
            
            layout(local_size_x=4, local_size_y=4, local_size_z=4) in;

            layout(rgba32f, binding=0) uniform image3D img_in;
            layout(rgba32f, binding=1) uniform image3D img_out;

            void main() {
                vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xyz));
                imageStore(img_out, ivec3(gl_LocalInvocationID.xyz), fragment);
            }
            """
        )
        tex_in = self.ctx.texture3d((4, 4, 4), 4, data=array('f', [v for v in range(4 * 4 * 4 * 4)]), dtype="f4")
        tex_out = self.ctx.texture3d((4, 4, 4), 4, dtype="f4")

        tex_in.bind_to_image(0, read=True, write=False)
        tex_out.bind_to_image(1, read=False, write=True)
        program.run(group_x=1)

        data_in = struct.unpack("256f", tex_in.read())
        data_out = struct.unpack("256f", tex_out.read())

        self.assertEqual(data_in, data_out)

    def test_texture_array_image(self):
        program = self.program = self.ctx.compute_shader(
            """
            #version 430
            
            layout(local_size_x=4, local_size_y=4, local_size_z=4) in;

            layout(rgba32f, binding=0) uniform image2DArray img_in;
            layout(rgba32f, binding=1) uniform image2DArray img_out;

            void main() {
                vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xyz));
                imageStore(img_out, ivec3(gl_LocalInvocationID.xyz), fragment);
            }
            """
        )
        tex_in = self.ctx.texture_array((4, 4, 4), 4, data=array('f', [v for v in range(4 * 4 * 4 * 4)]), dtype="f4")
        tex_out = self.ctx.texture_array((4, 4, 4), 4, dtype="f4")

        tex_in.bind_to_image(0, read=True, write=False)
        tex_out.bind_to_image(1, read=False, write=True)
        program.run(group_x=1)

        data_in = struct.unpack("256f", tex_in.read())
        data_out = struct.unpack("256f", tex_out.read())

        self.assertEqual(data_in, data_out)

    def test_texture_cube_image(self):
        program = self.ctx.compute_shader(
            """
            #version 450
            
            layout(local_size_x=4, local_size_y=4) in;

            layout(rgba8, binding=0) uniform imageCube img_in;
            layout(rgba8, binding=1) uniform imageCube img_out;

            void main() {
                for (int i = 0; i < 6; i++) {
                    vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xy, i));
                    imageStore(img_out, ivec3(gl_LocalInvocationID.xy, i), fragment);
                }
            }
            """
        )
        data = ([1] * 64) + ([2] * 64) + ([3] * 64) + ([4] * 64) + ([5] * 64) + ([6] * 64)
        tex_in = self.ctx.texture_cube((4, 4), 4, data=array('B', data), dtype="f1")
        tex_out = self.ctx.texture_cube((4, 4), 4, data=array('B', [0] * 64 * 6), dtype="f1")
        tex_in.bind_to_image(0, read=True, write=False)
        tex_out.bind_to_image(1, read=False, write=True)
        program.run(group_x=1)

        for face in range(0, 6):
            self.assertEqual(
                struct.unpack("64B", tex_in.read(face)),
                struct.unpack("64B", tex_out.read(face)),
            )


if __name__ == '__main__':
    unittest.main()



from array import array
import struct
import unittest

import moderngl
from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()
        cls.vbo = cls.ctx.buffer(reserve=1024)
        cls.res = cls.ctx.buffer(reserve=1024)
        cls.vbo.clear()

        cls.shader = '''
            #version %(version)s
            uniform %(type)s Uniform;
            in %(type)s v_in;
            out %(type)s v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        '''

    def test_float_uniform(self):
        prog = self.ctx.program(
            vertex_shader='''
            #version 330
            uniform float Uniform;
            in float v_in;
            out float v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''',
            varyings=['v_out']
        )
        vbo = self.ctx.buffer(struct.pack('f', 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, 'v_in')

        prog['Uniform'].value = 7.0
        vao.transform(self.res)

        val = struct.unpack('f', self.res.read(4))[0]
        self.assertAlmostEqual(val, 7.0)

    def test_int_uniform(self):
        prog = self.ctx.program(
            vertex_shader='''
                #version 330
                uniform int Uniform;
                in int v_in;
                out int v_out;
                void main() {
                    v_out = Uniform * v_in;
                }
            ''',
            varyings=['v_out']
        )
        vbo = self.ctx.buffer(struct.pack('i', 1))
        vao = self.ctx.simple_vertex_array(prog, vbo, 'v_in')

        prog['Uniform'].value = -2
        vao.transform(self.res)

        val = struct.unpack('i', self.res.read(4))[0]
        self.assertAlmostEqual(val, -2)

    def test_vec_uniform(self):
        prog = self.ctx.program(
            vertex_shader='''
                #version 330
                uniform vec3 Uniform;
                in vec3 v_in;
                out vec3 v_out;
                void main() {
                    v_out = Uniform * v_in;
                }
            ''',
            varyings=['v_out']
        )

        vbo = self.ctx.buffer(struct.pack('3f', 1.0, 1.0, 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, 'v_in')

        prog['Uniform'].value = (0.5, 1.0, 1.5)
        vao.transform(self.res)

        x, y, z = struct.unpack('3f', self.res.read(12))
        self.assertAlmostEqual(x, 0.5)
        self.assertAlmostEqual(y, 1.0)
        self.assertAlmostEqual(z, 1.5)

    def test_mat_uniform(self):
        prog = self.ctx.program(
            vertex_shader='''
                #version 330
                uniform mat2x3 Uniform;
                in float v_in;
                out mat2x3 v_out;
                void main() {
                    v_out = Uniform * v_in;
                }
            ''',
            varyings=['v_out']
        )

        vbo = self.ctx.buffer(struct.pack('f', 1.0))
        vao = self.ctx.simple_vertex_array(prog, vbo, 'v_in')

        prog['Uniform'].value = (0.0, 1.0, 2.0, 3.0, 4.0, 5.0)
        vao.transform(self.res)

        m = struct.unpack('6f', self.res.read(24))
        self.assertAlmostEqual(m[0], 0.0)
        self.assertAlmostEqual(m[1], 1.0)
        self.assertAlmostEqual(m[2], 2.0)
        self.assertAlmostEqual(m[3], 3.0)
        self.assertAlmostEqual(m[4], 4.0)
        self.assertAlmostEqual(m[5], 5.0)

    def test_sampler_2d(self):
        """RGBA8 2d sampler"""
        prog = self.ctx.program(vertex_shader="""
            #version 330
            uniform sampler2D tex;
            out float color;
            void main() {
                color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
            }
            """,
            varyings=['color',],
        )
        tex = self.ctx.texture((4, 1), 1, data=array('B', [127, 0, 255, 64]), dtype='f1')
        buff = self.ctx.buffer(reserve=4 * 4)
        vao = self.ctx.vertex_array(prog, [])
        tex.use(0)
        vao.transform(buff, vertices=4)
        data = struct.unpack('4f', buff.read())
        self.assertAlmostEqual(data[0], 0.498, places=2)
        self.assertAlmostEqual(data[1], 0.0)
        self.assertAlmostEqual(data[2], 1.0)
        self.assertAlmostEqual(data[3], 0.25, places=2)

    def test_sampler_2d_int(self):
        prog = self.ctx.program(vertex_shader="""
            #version 330
            uniform isampler2D tex;
            out int color;
            void main() {
                color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
            }
            """,
            varyings=['color',],
        )
        tex = self.ctx.texture((4, 1), 1, data=array('i', [-1, 0, 1000, 4353454]), dtype='i4')
        buff = self.ctx.buffer(reserve=4 * 4)
        vao = self.ctx.vertex_array(prog, [])
        tex.use(0)
        vao.transform(buff, vertices=4)
        data = struct.unpack('4i', buff.read())
        self.assertEqual(data[0], -1)
        self.assertEqual(data[1], 0)
        self.assertEqual(data[2], 1000)
        self.assertEqual(data[3], 4353454)

    def test_sampler_2d_uint(self):
        prog = self.ctx.program(vertex_shader="""
            #version 330
            uniform usampler2D tex;
            out uint color;
            void main() {
                color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
            }
            """,
            varyings=['color',],
        )
        tex = self.ctx.texture((4, 1), 1, data=array('I', [0, 500, 1000, 4353454]), dtype='u4')
        buff = self.ctx.buffer(reserve=4 * 4)
        vao = self.ctx.vertex_array(prog, [])
        tex.use(0)
        vao.transform(buff, vertices=4)
        data = struct.unpack('4I', buff.read())
        print(data)
        self.assertEqual(data[0], 0)
        self.assertEqual(data[1], 500)
        self.assertEqual(data[2], 1000)
        self.assertEqual(data[3], 4353454)

    def test_sampler_2d_array(self):
        """RGBA8 2d array sampler. Read two pixels from two different layers"""
        prog = self.ctx.program(vertex_shader="""
            #version 330
            uniform sampler2DArray tex;
            out float color;
            void main() {
                color = texelFetch(tex, ivec3(gl_VertexID % 2, 0, gl_VertexID / 2), 0).r;
            }
            """,
            varyings=['color',],
        )
        tex = self.ctx.texture_array((2, 1, 2), 1, data=array('B', [127, 0, 255, 64]), dtype='f1')
        buff = self.ctx.buffer(reserve=4 * 4)
        vao = self.ctx.vertex_array(prog, [])
        tex.use(0)
        vao.transform(buff, vertices=4)
        data = struct.unpack('4f', buff.read())
        self.assertAlmostEqual(data[0], 0.498, places=2)
        self.assertAlmostEqual(data[1], 0.0)
        self.assertAlmostEqual(data[2], 1.0)
        self.assertAlmostEqual(data[3], 0.25, places=2)

    def test_sampler_1d(self):
        prog = self.ctx.program(
            vertex_shader="""
            #version 330
            uniform sampler1D tex;
            out vec4 color;
            void main() {
                color = texelFetch(tex, 0, 0);
            }
            """,
            varyings=["color"],
        )
        self.assertIn("tex", prog)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import _example


class Example(_example.Example):
    title = 'Program Uniforms'
    window_size = (400, 400)
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform vec2 WindowSize;

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert / WindowSize * 2.0 - 1.0, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 Color;

                out vec4 f_color;

                void main() {
                    f_color = vec4(Color, 1.0);
                }
            ''',
        )

        self.prog['WindowSize'] = self.window_size
        self.prog['Color'] = (0.2, 0.4, 0.7)

        vertices = np.array([
            100.0, 100.0,
            300.0, 200.0,
            200.0, 300.0,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time: float, frame_time: float):
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.vao.render()


if __name__ == '__main__':
    Example.run()



"""
Simple test script for gc

* Leak then mgl object fails on creation?
* If object doesn't have ctx attribute it's not valid?
"""
import logging
import time
from ctypes import c_long

import numpy as np
import moderngl

logging.basicConfig(level=logging.DEBUG)

ctx = moderngl.create_context(standalone=True, require=330)

print(ctx)
ctx.gc_mode = "auto"


def ref_count():
    """Simple refcount demo"""
    var1 = [1, 2]
    print(c_long.from_address(id(var1)).value)


# --- Texture ---

def test_texture():
    return ctx.texture((100, 100), 4, data=np.random.randint(0, 1, 100 * 100 * 4, dtype="u1"), dtype="u1")

def test_texture_mass_create(n=1_000):
    for _ in range(n):
        texture = ctx.texture((1000, 1000), 4, data=np.random.randint(0, 255, 1000 * 1000 * 4, dtype="u1"))


def test_texture_failure():
    try:
        texture = ctx.texture((1000, 1000), 4, data=np.random.randint(0, 255, 1000 * 1001 * 4, dtype="u1"))
    except Exception as ex:
        print(ex)


def test_texture_failure_mass_create(n=1000):
    for i in range(n):
        test_texture_failure()


# --- TextureArray ---

def test_texture_array():
    texture = ctx.texture_array((100, 100, 100), 4, data=np.random.randint(0, 255, 100 * 100 * 100 * 4, dtype="u1"))


def test_texture_array_mass_create(n=1000):
    for _ in range(n):
        texture = ctx.texture_array((100, 100, 100), 4, data=np.random.randint(0, 255, 100 * 100 * 100 * 4, dtype="u1"))


def test_texture_array_failure():
    try:
        texture = ctx.texture_array((100, 100, 10), 4, data=np.random.randint(0, 255, 100 * 100 * 100 * 4, dtype="u1"))
    except Exception as ex:
        print(ex)


def test_texture_array_failure_mass_create(n=1000):
    for i in range(n):
        test_texture_array_failure()


# --- Texture3D ---

def test_texture_3d():
    texture = ctx.texture3d((100, 100, 100), 4, data=np.random.randint(0, 255, 100 * 100 * 100 * 4, dtype="u1"))


def test_texture_3d_mass_create(n=1000):
    for _ in range(n):
        texture = ctx.texture3d((100, 100, 100), 4, data=np.random.randint(0, 255, 100 * 100 * 100 * 4, dtype="u1"))


def test_texture_3d_failure():
    try:
        texture = ctx.texture3d((100, 100, 100), 4, data=np.random.randint(0, 255, 100 * 100 * 101 * 4, dtype="u1"))
    except Exception as ex:
        print(ex)


def test_texture_3d_failure_mass_create(n=1000):
    for i in range(n):
        test_texture_3d_failure()


# --- Framebuffer ---

def test_framebuffer():
    texture = ctx.texture((100, 100), 4)
    fbo = ctx.framebuffer(color_attachments=[texture])


def test_framebuffer_mass_create(n=1000):
    for i in range(n):
        texture = ctx.texture((100, 100), 4)
        ctx.framebuffer(color_attachments=[texture])
        # Depending on drivers FBOs might not release until flush/finish
        ctx.finish()


def test_framebuffer_failure():
    try:
        fbo = ctx.framebuffer(color_attachments=[])
    except Exception as ex:
        print(ex)


def test_framebuffer_failure_mass_create(n=1000):
    for i in range(n):
        test_framebuffer_failure()


# --- RenderBuffer ---

def test_renderbuffer_mass_create(n=1000):
    for i in range(n):
        rb = ctx.renderbuffer((1000, 1000))
        ctx.finish()


def test_renderbuffer_failure_mass_create(n=1000):
    for i in range(n):
        try:
            rb = ctx.renderbuffer((1000, 1000), components=4, dtype="moo")
        except Exception as ex:
            print(ex)
        ctx.finish()

# --- Buffer ---

def test_buffer():
    buffer = ctx.buffer(data=np.random.randint(0, 255, 1024 * 1024 * 1024, dtype="u1"))
    buffer.release()


def test_buffer_mass_create(n=1000):
    for i in range(n):
        b = ctx.buffer(data=np.random.randint(0, 255, 1024 * 1024, dtype="u1"))


def test_buffer_creation_failed():
    try:
        buffer = ctx.buffer(data=None)
    except Exception as ex:
        print(ex)


def test_buffer_creation_failed_mass_create(n=1000):
    for i in range(n):
        test_buffer_creation_failed()


# --- Query ---

def test_query_mass_create(n=1000):
    for _ in range(n):
        q = ctx.query(samples=True, any_samples=True, time=True, primitives=True)
        # print(c_long.from_address(id(q.mglo)))


# --- VertexArray / Program ---

prog_src = {
        "vertex_shader": """
        #version 330
        in vec2 in_pos;
        void main() {
            gl_Position = vec4(in_pos, 0.0, 1.0);
        }
        """,
        "fragment_shader": """
        #version 330
        out vec4 color;
        void main() {
            color = vec4(1.0);
        }
        """
}


def test_vertex_array():
    buffer = ctx.buffer(reserve=1024)
    prog = ctx.program(**prog_src)
    vao = ctx.vertex_array(prog, [(buffer, "2f", "in_pos")])


def test_vertex_array_mass_create(n=1000):
    for i in range(n):
        prog = ctx.program(**prog_src)
        vbo = ctx.buffer(reserve=1_000_000)
        ibo = ctx.buffer(reserve=1_000_000)
        vao = ctx.vertex_array(prog, [(vbo, "2f", "in_pos")], index_buffer=ibo)


def test_vertex_array_creation_failure():
    try:
        buffer = ctx.buffer(reserve=1024)
        prog = ctx.program(**prog_src)
        vao = ctx.vertex_array(prog, [(buffer, "2f", "in_pos_invalid")])
    except moderngl.error.Error:
        pass


def test_vertex_array_creation_failure_mass_create(n=1000):
    for i in range(n):
        test_vertex_array_creation_failure()


# --- ComputeShader ---

def test_compute_shader():
    compute_shader = ctx.compute_shader('''
        #version 430
        layout (local_size_x = 1, local_size_y = 1, local_size_z = 1) in;
        void main() {
        }
    ''')


def test_compute_shader_mass_create(n=1000):
    for i in range(n):
        test_compute_shader()
        print(i)


def test_compute_shader_fail():
    try:
        compute_shader = ctx.compute_shader('''
            #version 430
            layout (local_size_x = 1, local_size_y = 1, local_size_z = 1) in;
            void main() {
            }
        ''')
    except Exception as ex:
        print(ex)


def test_compute_shader_fail_mass_create(n=1000):
    for i in range(n):
        test_compute_shader_fail()


# --- Sampler ---

def test_sampler_mass_create(n=1000):
    for i in range(n):
        sampler = ctx.sampler()


# --- Scope ---

def test_scope():
    fbo = ctx.simple_framebuffer((100, 100))
    tex1 = ctx.texture((10, 100), 4)
    tex2 = ctx.texture((10, 100), 4)
    ubo1 = ctx.buffer(reserve=16000)
    ubo2 = ctx.buffer(reserve=16000)
    ssob1 = ctx.buffer(reserve=1024)
    ssob2 = ctx.buffer(reserve=1024)

    for i in range(10_000):
        print(i)
        scope = ctx.scope(
            framebuffer=fbo,
            enable_only=moderngl.NOTHING,
            textures=[(tex1, 0), (tex2, 1)],
            uniform_buffers=[(ubo1, 0), (ubo2, 1)],
            storage_buffers=[(ssob1, 0), (ssob2, 1)],
        )
        # scope = ctx.scope(
        #     framebuffer=ctx.simple_framebuffer((100, 100)),
        #     enable_only=moderngl.NOTHING,
        #     textures=[(ctx.texture((10, 100), 4), 0), (ctx.texture((10, 100), 4), 1)],
        #     uniform_buffers=[(ctx.buffer(reserve=16000), 0), (ctx.buffer(reserve=16000), 1)],
        #     storage_buffers=[(ctx.buffer(reserve=1024), 0), (ctx.buffer(reserve=1024), 1)],
        # )
        # with scope:
        #     pass


# ------------------------------------------------------------------------------------------

# ref_count()

N = 100

test_texture()
test_texture_mass_create(N)
test_texture_failure()
test_texture_failure_mass_create(N)

test_texture_array()
test_texture_array_mass_create(N)
test_texture_array_failure()
test_texture_array_failure_mass_create(N)

test_texture_3d()
test_texture_3d_mass_create(N)
test_texture_3d_failure()
test_texture_3d_failure_mass_create(N)

test_framebuffer()
test_framebuffer_mass_create(N)
test_framebuffer_failure()
test_framebuffer_failure_mass_create(N)

test_renderbuffer_mass_create(N)
test_renderbuffer_failure_mass_create(N)

test_buffer()
test_buffer_mass_create(N)
test_buffer_creation_failed()
test_buffer_creation_failed_mass_create(N)

test_query_mass_create()

test_vertex_array()
test_vertex_array_mass_create(N)
test_vertex_array_creation_failure()
test_vertex_array_creation_failure_mass_create(N)

if ctx.version_code >= 430:
    test_compute_shader()
    test_compute_shader_mass_create(N)
    test_compute_shader_fail()
    test_compute_shader_fail_mass_create(N)

test_sampler_mass_create(N)

# test_scope()
print("Done. Waiting")
time.sleep(300)



import os
import unittest

import moderngl


class TestCase(unittest.TestCase):

    def test_local(self):
        path1 = os.path.normpath(os.path.dirname(os.path.dirname(moderngl.__file__)))
        path2 = os.path.normpath(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        self.assertEqual(path1.lower(), path2.lower(), msg='Using installed moderngl')


if __name__ == '__main__':
    unittest.main()



"""
Using a subprocess in python to calculate data on a separate core.
Threads in python are not really suitable for this task because
of the GIL and we would just be slowing down the main thread.

This is a fairly simple example spawning a deamon thread that
throws new random textures on a queue that can be read back by
the main process.

There are probably better ways of doing this, but this is a good start.
For example: We might want spawn multiple processes that writes to
the same queue.
"""
from time import sleep
import multiprocessing as mp
from queue import Empty
import struct
import random
import moderngl_window
from moderngl_window import geometry
from pyrr import matrix44


class SubprocessTest(moderngl_window.WindowConfig):
    window_size = 512, 512
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.texture = self.ctx.texture((100, 100), 3, data=gen_texture())
        self.texture.filter = self.ctx.NEAREST, self.ctx.NEAREST
        self.quad = geometry.quad_2d(size=(1.3, 1.3))
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330
            uniform mat4 model;
            in vec2 in_position;
            in vec2 in_texcoord_0;
            out vec2 uv;
            void main () {
                gl_Position = model * vec4(in_position, 0.0, 1.0);
                uv = in_texcoord_0;
            }
            """,
            fragment_shader="""
            #version 330
            uniform sampler2D tex;
            in vec2 uv;
            out vec4 f_color;
            void main() {
                f_color = texture(tex, uv);
            }
            """,
        )
        self.generator = TextureGenerator()
        self.generator.start()

    def next_texture(self):
        """Attempt to get a new texture from the queue"""
        try:
            data = self.generator.queue.get(block=False)
            self.texture.write(data)
        except Empty:
            pass

    def render(self, time: float, frame_time: float):
        self.ctx.clear()
        self.texture.use()
        self.program["model"].write(matrix44.create_from_axis_rotation([0.0, 0.0, 1.0], time, dtype="f4"))
        self.quad.render(self.program)

        self.next_texture()

    def close(self):
        self.generator.stop()


class TextureGenerator:

    def __init__(self):
        self.queue = mp.Queue()
        self.process = mp.Process(
            target=self.do_work,
            args=(self.queue,),
            daemon=True,
        )

    def do_work(self, queue):
        while True:
            # Back off a little bit to avoid overloading the queue
            if self.queue.qsize() > 60:
                sleep(0.5)

            queue.put(gen_texture(), block=False)

    def start(self):
        self.process.start()

    def stop(self):
        self.process.terminate()


def gen_texture():
    """
    An inefficient way of generating random texture.
    The goal is to make the process reasonably busy.
    """
    num_frags = 100 * 100 * 3
    data = [random.randint(0, 255) for _ in range(num_frags)]
    return struct.pack(f"{num_frags}B", *data)


if __name__ == "__main__":
    SubprocessTest.run()



"""
Shows transform feedback with geometry shaders were
the input and output primitive is different.

We have a list of points and emit triangle data to a buffer
It creates a quads per point using two triangles.
"""
import struct
from ported._example import Example
import moderngl


class GenerateData(Example):
    title = "Transform Feedback Geometry shader"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330
                in vec3 in_vert;
                void main() {
                    gl_Position = vec4(in_vert, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330
                out vec4 f_color;
                void main() {
                    f_color = vec4(1.0);
                }
            ''',
        )

        # Transform program with geo shader
        self.transform_prog = self.ctx.program(
            vertex_shader='''
                #version 330

                void main() {
                    gl_Position = vec4(gl_VertexID/5.0 - 0.9, 0.0, 0.0, 1.0);
                }
            ''',
            geometry_shader="""
            #version 330
            layout(points) in;
            layout(triangle_strip, max_vertices = 6) out;
            out vec3 pos;
            const float SIZE = 0.05;
            void main() {
                vec3 center = gl_in[0].gl_Position.xyz;

                // First triangle
                pos = center + vec3(-SIZE, SIZE, 0.0);
                EmitVertex();
                pos = center + vec3(-SIZE, -SIZE, 0.0);
                EmitVertex();
                pos = center + vec3(SIZE, SIZE, 0.0);
                EmitVertex();
                EndPrimitive();

                // Second triangle
                pos = center + vec3(SIZE, SIZE, 0.0);
                EmitVertex();
                pos = center + vec3(-SIZE, -SIZE, 0.0);
                EmitVertex();
                pos = center + vec3(SIZE, -SIZE, 0.0);
                EmitVertex();

                EndPrimitive();
            }
            """,
            varyings=['pos'],
        )

        N = 10
        self.buffer = self.ctx.buffer(reserve=N * 12 * 6)  # N * 6 x vec3
        self.transform_vao = self.ctx.vertex_array(self.transform_prog, [])
        self.transform_vao.transform(self.buffer, mode=moderngl.POINTS, vertices=N)

        self.vao = self.ctx.vertex_array(
            self.prog,
            [(self.buffer, '3f', 'in_vert')]
        )

    def render(self, time, frame_time):
        self.wnd.clear()
        self.vao.render(mode=moderngl.TRIANGLES)


if __name__ == '__main__':
    GenerateData.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.3, 0.5, 1.0, 1.0);
#         }
#     '''),
# ])

# vbo = ctx.buffer(struct.pack(
#     '12f',
#     -0.5, -0.5,
#     -0.5, 0.5,
#     0.5, -0.5,
#     -0.5, 0.5,
#     0.5, -0.5,
#     0.5, 0.5,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# ctx.enable(ModernGL.CULL_FACE)

# while wnd.update():
#     if wnd.time % 1.0 < 0.5:
#         ctx.front_face = 'CW'
#     else:
#         ctx.front_face = 'CCW'

#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()



import moderngl
import numpy as np

from PIL import Image

ctx = moderngl.create_standalone_context()

prog = ctx.program(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec3 in_color;

        out vec3 v_color;

        void main() {
            v_color = in_color;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330

        in vec3 v_color;

        out vec3 f_color;

        void main() {
            f_color = v_color;
        }
    ''',
)

x = np.linspace(-1.0, 1.0, 50)
y = np.random.rand(50) - 0.5
r = np.ones(50)
g = np.zeros(50)
b = np.zeros(50)

vertices = np.dstack([x, y, r, g, b])

vbo = ctx.buffer(vertices.astype('f4').tobytes())
vao = ctx.simple_vertex_array(prog, vbo, 'in_vert', 'in_color')

fbo = ctx.simple_framebuffer((512, 512))
fbo.use()
fbo.clear(0.0, 0.0, 0.0, 1.0)
vao.render(moderngl.LINE_STRIP)

Image.frombytes('RGB', fbo.size, fbo.read(), 'raw', 'RGB', 0, -1).show()



import numpy as np

import moderngl
from ported._example import Example


class Fractal(Example):
    title = "Julia Fractal"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 v_text;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                    v_text = in_vert / 2.0 + 0.5;
                }
            ''',
            fragment_shader='''
                #version 330

                in vec2 v_text;
                out vec4 f_color;

                uniform vec2 Center;
                uniform int Iter;

                void main() {
                    vec2 z = vec2(5.0 * (v_text.x - 0.5), 3.0 * (v_text.y - 0.5));
                    vec2 c = Center;

                    int i;
                    for(i = 0; i < Iter; i++) {
                        vec2 v = vec2(
                            (z.x * z.x - z.y * z.y) + c.x,
                            (z.y * z.x + z.x * z.y) + c.y
                        );
                        if (dot(v, v) > 4.0) break;
                        z = v;
                    }

                    float cm = fract((i == Iter ? 0.0 : float(i)) * 10 / Iter);
                    f_color = vec4(
                        fract(cm + 0.0 / 3.0),
                        fract(cm + 1.0 / 3.0),
                        fract(cm + 2.0 / 3.0),
                        1.0
                    );
                }
            ''',
        )

        self.center = self.prog['Center']
        self.iter = self.prog['Iter']

        vertices = np.array([-1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0])

        self.vbo = self.ctx.buffer(vertices.astype('f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)

        self.center.value = (0.49, 0.32)
        self.iter.value = 100

        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    Fractal.run()



import numpy as np

import _example


class Example(_example.Example):
    title = 'Alpha Blending'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                in vec4 in_color;

                out vec4 v_color;

                uniform vec2 Scale;
                uniform float Rotation;

                void main() {
                    v_color = in_color;
                    float r = Rotation * (0.5 + gl_InstanceID * 0.05);
                    mat2 rot = mat2(cos(r), sin(r), -sin(r), cos(r));
                    gl_Position = vec4((rot * in_vert) * Scale, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330
                in vec4 v_color;
                out vec4 f_color;
                void main() {
                    f_color = vec4(v_color);
                }
            ''',
        )

        self.prog['Scale'] = (0.5, self.aspect_ratio * 0.5)

        vertices = np.array([
            1.0, 0.0,
            1.0, 0.0, 0.0, 0.5,

            -0.5, 0.86,
            0.0, 1.0, 0.0, 0.5,

            -0.5, -0.86,
            0.0, 0.0, 1.0, 0.5,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert', 'in_color')
        self.vao.scope = self.ctx.scope(enable=self.ctx.BLEND)
        self.vao.instances = 10

    def render(self, time: float, frame_time: float):
        self.prog['Rotation'] = time
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.vao.render()


if __name__ == '__main__':
    Example.run()



"""
Quick and dirty example showing how sprites can be rendered using a geometry shader.
We also show simple scrolling with projection matrix.

The goal is to redice the sprite data on the client as much as possible.
We can define a sprite by its position, size and rotation. This can be
expressed in 5 32 bit floats. This also opens up for individually rotating
each sprite in the shader itself. This technique can be extended with more
sprite parameters.

Other optimizations that can be done:
* Cull sprites outside the viewport in geo shader
"""
import math

import moderngl
from ported._example import Example

from pyrr import Matrix44
from array import array


class GeoSprites(Example):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.ball_texture = self.load_texture_2d("ball.png")

        # Sprite shader using geometry shader
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            // The per sprite input data
            in vec2 in_position;
            in vec2 in_size;
            in float in_rotation;

            out vec2 size;
            out float rotation;

            void main() {
                // We just pass the values unmodified to the geometry shader
                gl_Position = vec4(in_position, 0, 1);
                size = in_size;
                rotation = in_rotation;
            }
            """,
            geometry_shader="""
            #version 330

            // We are taking single points form the vertex shader
            // and emitting 4 new vertices creating a quad/sprites
            layout (points) in;
            layout (triangle_strip, max_vertices = 4) out;

            uniform mat4 projection;

            // Since geometry shader can take multiple values from a vertex
            // shader we need to define the inputs from it as arrays.
            // In our instance we just take single values (points)
            in vec2 size[];
            in float rotation[];
            out vec2 uv;

            void main() {
                // We grab the position value from the vertex shader
                vec2 center = gl_in[0].gl_Position.xy;
                // Calculate the half size of the sprites for easier calculations
                vec2 hsize = size[0] / 2.0;
                // Convert the rotation to radians
                float angle = radians(rotation[0]);

                // Create a 2d rotation matrix
                mat2 rot = mat2(
                    cos(angle), sin(angle),
                    -sin(angle), cos(angle)
                );

                // Emit a triangle strip creating a quad (4 vertices).
                // Here we need to make sure the rotation is applied before we position the sprite.
                // We just use hardcoded texture coordinates here. If an atlas is used we
                // can pass an additional vec4 for specific texture coordinates.

                // Each EmitVertex() emits values down the shader pipeline just like a single
                // run of a vertex shader, but in geomtry shaders we can do it multiple times!

                // Upper left
                gl_Position = projection * vec4(rot * vec2(-hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 1);
                EmitVertex();

                // lower left
                gl_Position = projection * vec4(rot * vec2(-hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 0);
                EmitVertex();

                // upper right
                gl_Position = projection * vec4(rot * vec2(hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 1);
                EmitVertex();

                // lower right
                gl_Position = projection * vec4(rot * vec2(hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 0);
                EmitVertex();

                // We are done with this triangle strip now
                EndPrimitive();
            }

            """,
            fragment_shader="""
            #version 330

            uniform sampler2D sprite_texture;

            in vec2 uv;
            out vec4 fragColor;
            
            void main() {
                fragColor = texture(sprite_texture, uv);
            }
            """,
        )

        self.sprite_data_size = 5 * 4  # 5 32 bit floats
        self.sprite_data = self.ctx.buffer(reserve=1000 * self.sprite_data_size)  # Capacity for 1000 sprites
        self.vao = self.ctx.vertex_array(
            self.program,
            [
                (self.sprite_data, "2f 2f 1f", "in_position", "in_size", "in_rotation"),
            ]
        )

    def render(self, time, frame_time):
        self.ctx.clear()
        self.ctx.enable(moderngl.BLEND)

        num_sprites = 16
        # We'll just generate some sprite data on the fly here.
        # This should only be necessary every time the sprite data changes (in a prefect wold)

        # Grab the size of the screen or current render target
        width, height = self.ctx.fbo.size

        # We just create a generator function instead of 
        def gen_sprites(time):
            rot_step = math.pi * 2 / num_sprites
            for i in range(num_sprites):
                # Position
                yield width / 2 + math.sin(time + rot_step * i) * 300
                yield height / 2 + math.cos(time + rot_step * i) * 300
                # size
                yield 100
                yield 100
                # rotation
                yield math.sin(time + i) * 200

        self.sprite_data.write(array("f", gen_sprites(time)))

        # calculate scroll offsets. We truncate to integers here.
        # This depends what "look" you want for your game.
        scroll_x, scroll_y = int(math.sin(time) * 100), int(math.cos(time) * 100)

        # Create a orthogonal projection matrix
        # Let's also modify the projection to scroll the entire screen.
        projection = Matrix44.orthogonal_projection(
            scroll_x,   # left
            width + scroll_x,   # right
            height + scroll_y,   # top
            scroll_y,   # bottom
            1,   # near
            -1,  # far
            dtype="f4",  # ensure we create 32 bit value (64 bit is default)
        )

        # Update the projection uniform
        self.program["projection"].write(projection)
        # Configure the sprite_texture uniform to read from texture channel 0
        self.program["sprite_texture"] = 0
        # Bind the texture to channel 0
        self.ball_texture.use(0)
        # Since we have overallocated the buffer (room for 1000 sprites) we 
        # need to specify how many we actually want to render passing number of vertices.
        # Also the mode needs to be the same as the geometry shader input type (points!)
        self.vao.render(mode=moderngl.POINTS, vertices=num_sprites)


if __name__ == "__main__":
    GeoSprites.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# # Shaders & Program

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.3, 0.5, 1.0, 1.0);
#         }
#     '''),
# ])

# # Buffer

# vbo = ctx.buffer(struct.pack(
#     '6f',
#     0.0, 0.8,
#     -0.6, -0.8,
#     0.6, -0.8,
# ))

# # Put everything together

# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# # Main loop

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_sampler_2D_getsetter(self):
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330
                uniform sampler2D Texture2D;
                void main() {
                    gl_Position = texture(Texture2D, vec2(0.0, 0.0));
                }
            '''
        )

        sampler_uniform = self.prog['Texture2D']

        sampler_uniform.value = 3
        self.assertEqual(sampler_uniform.value, 3)

        sampler_uniform.value = 1
        self.assertEqual(sampler_uniform.value, 1)

    def test_sampler_3D_getsetter(self):
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330
                uniform sampler3D Texture3D;
                void main() {
                    gl_Position = texture(Texture3D, vec3(0.0, 0.0, 0.0));
                }
            '''
        )

        sampler_uniform = self.prog['Texture3D']

        sampler_uniform.value = 3
        self.assertEqual(sampler_uniform.value, 3)

        sampler_uniform.value = 1
        self.assertEqual(sampler_uniform.value, 1)


if __name__ == '__main__':
    unittest.main()



import unittest

import moderngl

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_create(self):
        size = (64, 32, 8)
        texture = self.ctx.texture_array(size, 4)

        # Basic attributes
        assert texture.size == size
        assert texture.width == 64
        assert texture.height == 32
        assert texture.layers == 8
        assert texture.dtype == 'f1'
        assert texture.components == 4

        # Texture paramters
        assert texture.repeat_x == True
        assert texture.repeat_y == True
        assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)
        assert texture.swizzle == "RGBA"
        print(texture.anisotropy)
        assert texture.anisotropy == 1.0

        texture.build_mipmaps()
        assert texture.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)

    def test_texture_default_filter(self):
        """Ensure default filter is correct"""
        # Float types
        for dtype in ["f1", "f2", "f4"]:
            texture = self.ctx.texture_array((10, 10, 10), 4, dtype=dtype)
            self.assertEqual(texture.filter, (moderngl.LINEAR, moderngl.LINEAR))

        for dtype in ["u1", "u2", "u4", "i1", "i2", "i4"]:
            texture = self.ctx.texture_array((10, 10, 10), 4, dtype=dtype)
            self.assertEqual(texture.filter, (moderngl.NEAREST, moderngl.NEAREST))



import unittest

from common import get_context


class TestBuffer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        buf1 = self.ctx.buffer(data=b'\xAA\x55' * 10)
        buf2 = self.ctx.buffer(reserve=buf1.size)
        buf2.write(b'Hello World!')

        self.assertEqual(buf1.read(5, offset=1), b'\x55\xaa\x55\xaa\x55')
        self.assertEqual(buf2.read(5, offset=6), b'World')

        self.ctx.copy_buffer(buf2, buf1, read_offset=1, write_offset=6, size=5)
        self.assertEqual(buf2.read(12), b'Hello \x55\xaa\x55\xaa\x55!')


if __name__ == '__main__':
    unittest.main()



"""
Links:
- http://www.opengl-tutorial.org/fr/intermediate-tutorials/tutorial-16-shadow-mapping/#shadowmap-basique
- https://learnopengl.com/Advanced-Lighting/Shadows/Shadow-Mapping
- https://learnopengl.com/Advanced-OpenGL/Depth-testing
"""
from typing import Final, Dict

import numpy as np
from moderngl_window.scene import Scene
from pyrr import Matrix44
from pyrr import Vector3

import moderngl
#
from ported._example import Example

SHADOW_SIZE: Final[int] = 2 << 7  # 512²


class ShadowMappingSample(Example):
    title = "ShadowMapping"
    window_size = (1280, 720)
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog_render_scene_with_sm = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 u_mvp;
                uniform mat4 u_depth_bias_mvp;

                in vec3 in_position;
                in vec3 in_normal;
                in vec2 in_texcoord_0;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_text;

                out vec4 v_shadow_coord;

                void main() {
                    gl_Position = u_mvp * vec4(in_position, 1.0);

                    v_shadow_coord = u_depth_bias_mvp * vec4(in_position, 1.0);

                    v_vert = in_position;
                    v_norm = in_normal;
                    v_text = in_texcoord_0;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 u_light;
                uniform vec3 u_color;

                uniform sampler2DShadow u_sampler_shadow;
                
                uniform bool u_use_color_texture;
                uniform sampler2D u_sampler_billboard;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_text;

                in vec4 v_shadow_coord;

                out vec4 f_color;
                
                const float LIGHT_AMBIENT = 0.20;

                float compute_visibility(in float cos_theta) {
                    // shadow coordinates in light space
                    vec2 shadow_coord_ls = v_shadow_coord.xy / v_shadow_coord.w;
                    
                    // bias according to the slope
                    float bias = 0.005 * tan(acos(cos_theta));
                    bias = clamp(bias, 0, 0.01);

                    float z_from_cam = v_shadow_coord.z / v_shadow_coord.w - bias;
                    vec3 shadow_coord = vec3(shadow_coord_ls, z_from_cam);
                    float shadow_value = texture(u_sampler_shadow, shadow_coord);
                    return 1.0 - shadow_value;
                }
                                
                void main() {
                    // Lighting
                    // Diffuse lighting + ambient
                    vec3 light_vector_obj_space = normalize(u_light - v_vert);
                    vec3 normal_obj_space = normalize(v_norm);
                    float cos_theta = dot(light_vector_obj_space, normal_obj_space);
                    float diffuse = clamp(cos_theta, 0.0, 1.0);
                    // Shadow component                
                    float lum = mix(LIGHT_AMBIENT, 1.0, diffuse * compute_visibility(cos_theta));
                    // Color object (color or texture)                    
                    vec4 object_color = u_use_color_texture ? vec4(texture(u_sampler_billboard, v_text).rgb, 1.0) : vec4(u_color, 1.0);
                    // Final pixel color
                    f_color = object_color * lum;
                }
            ''',
        )

        self.prog_depth = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 u_mvp;

                in vec3 in_position;

                void main() {
                    gl_Position = u_mvp * vec4(in_position, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330
                void main() {
                }
            ''',
        )
        # textures unit
        self.prog_render_scene_with_sm['u_sampler_shadow'].value = 0
        self.prog_render_scene_with_sm['u_sampler_billboard'].value = 1
        # uniforms
        self.mvp = self.prog_render_scene_with_sm['u_mvp']
        self.mvp_depth = self.prog_render_scene_with_sm['u_depth_bias_mvp']
        self.light = self.prog_render_scene_with_sm['u_light']
        self.color = self.prog_render_scene_with_sm['u_color']
        self.use_color_texture = self.prog_render_scene_with_sm['u_use_color_texture']
        self.mvp_shadow = self.prog_depth['u_mvp']

        self.objects: Dict[str, moderngl.VertexArray] = {}
        self.objects_shadow: Dict[str, moderngl.VertexArray] = {}

        for name in {'ground', 'grass', 'billboard', 'billboard-holder', 'billboard-image'}:
            scene: Scene = self.load_scene(f'scene-1-{name}.obj')
            vao = scene.root_nodes[0].mesh.vao
            self.objects[name] = vao.instance(self.prog_render_scene_with_sm)
            self.objects_shadow[name] = vao.instance(self.prog_depth)

        # texture on billboard
        self.texture_on_billboard = self.load_texture_2d('infographic-1.jpg')
        self.texture_on_billboard.use(location=1)

        shadow_size = (SHADOW_SIZE, SHADOW_SIZE,)

        self.tex_depth = self.ctx.depth_texture(shadow_size)
        self.tex_color_depth = self.ctx.texture(shadow_size, components=1, dtype='f4')
        self.fbo_depth = self.ctx.framebuffer(
            color_attachments=[self.tex_color_depth],
            depth_attachment=self.tex_depth
        )
        self.sampler_depth = self.ctx.sampler(
            # bi-linear interpolation on depth fetch -> Percentage Close Filtering (2x2)
            filter=(moderngl.LINEAR, moderngl.LINEAR),
            compare_func='>=',  # enable depth func
            repeat_x=False, repeat_y=False,
        )

        self.ctx.enable(moderngl.CULL_FACE)

    def render(self, time: float, _frame_time: float):
        # pass 0: clear buffers
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        cam_proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        cam_look_at = Matrix44.look_at(
            (47.697, -8.147, 24.498),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 1.0),
        )
        cam_rotate = Matrix44.from_z_rotation(np.sin(time) * 0.5 + 0.2)
        cam_mvp = cam_proj * cam_look_at * cam_rotate
        self.mvp.write(cam_mvp.astype('f4').tobytes())

        # build light camera
        light_rotate = Matrix44.from_z_rotation(time)
        light_pos = light_rotate * Vector3((-60.69, -40.14, 52.49))
        self.light.value = tuple(light_pos)
        light_look_at = Matrix44.look_at(
            light_pos,
            target=(0, 0, 0),
            up=(0.0, 0.0, 1.0),
        )
        # light projection matrix (scene dependent)
        light_proj = Matrix44.perspective_projection(
            fovy=90.0 / 4,  # smaller value increase shadow precision
            aspect=self.wnd.aspect_ratio,
            near=60.0,
            far=100.0
        )
        # light model view projection matrix
        mvp_light = light_proj * light_look_at
        bias_matrix = (
                Matrix44.from_translation((0.5, 0.5, 0.5), dtype='f4') *
                Matrix44.from_scale((0.5, 0.5, 0.5), dtype='f4')
        )
        mvp_depth_bias = bias_matrix * mvp_light
        # send uniforms to shaders
        self.mvp_depth.write(mvp_depth_bias.astype('f4').tobytes())
        self.mvp_shadow.write(mvp_light.astype('f4').tobytes())

        # pass 1: render shadow-map (depth framebuffer -> texture) from light view
        self.fbo_depth.use()
        self.fbo_depth.clear(1.0, 1.0, 1.0)
        # https://moderngl.readthedocs.io/en/stable/reference/context.html?highlight=culling#moderngl.Context.front_face
        # clock wise -> render back faces
        self.ctx.front_face = 'cw'
        for vao_shadow in self.objects_shadow.values():
            vao_shadow.render()

        # pass 2: render the scene and retro project depth shadow-map
        # counter clock wise -> render front faces
        self.ctx.front_face = 'ccw'
        self.use_color_texture.value = False

        self.ctx.screen.use()
        self.sampler_depth.use(location=0)
        self.tex_depth.use(location=0)

        # pass 2a: render colored scene with shadow
        for object_name, object_color in (
                ("ground", (0.67, 0.49, 0.29)),
                ("grass", (0.46, 0.67, 0.29)),
                ("billboard", (1.0, 1.0, 1.0)),
                ("billboard-holder", (0.2, 0.2, 0.2)),
        ):
            self.color.value = object_color
            self.objects[object_name].render()
        # pass 2b: render textured scene with shadow
        self.use_color_texture.value = True
        self.objects['billboard-image'].render()


if __name__ == '__main__':
    ShadowMappingSample.run()



import unittest

from common import get_context


class TestBuffer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_buffer_clear_1(self):
        buf = self.ctx.buffer(data=b'\xAA\x55' * 10)
        buf.clear(chunk=b'AB')
        self.assertEqual(buf.read(), b'AB' * 10)

    def test_buffer_clear_2(self):
        buf = self.ctx.buffer(data=b'\xAA\x55' * 10)
        buf.clear(offset=1, size=18, chunk=b'AB')
        self.assertEqual(buf.read(), b'\xAAABABABABABABABABAB\x55')

    def test_buffer_create(self):
        buf = self.ctx.buffer(data=b'\xAA\x55' * 10)
        self.assertEqual(buf.read(), b'\xAA\x55' * 10)

    def test_buffer_read_write(self):
        buf = self.ctx.buffer(reserve=10)
        buf.write(b'abcd')
        self.assertEqual(buf.read(4), b'abcd')
        buf.write(b'abcd', offset=3)
        self.assertEqual(buf.read(4, offset=1), b'bcab')
        buf.write(b'abcd', offset=6)
        self.assertEqual(buf.read(), b'abcabcabcd')
        self.assertEqual(buf.read(offset=3), b'abcabcd')

    def test_buffer_read_into_1(self):
        data = b'Hello World!'
        buf = self.ctx.buffer(data)
        res = bytearray(buf.size)
        buf.read_into(res)
        self.assertEqual(bytes(res), data)

    def test_buffer_read_into_2(self):
        data = b'Hello World!'
        buf = self.ctx.buffer(data)
        res = bytearray(10)
        buf.read_into(res, offset=6, size=5, write_offset=0)
        buf.read_into(res, offset=0, size=5, write_offset=5)
        self.assertEqual(bytes(res), b'WorldHello')

    def test_buffer_orphan(self):
        buf = self.ctx.buffer(reserve=1024)
        buf.orphan()

    def test_buffer_orphan_resize(self):
        buf = self.ctx.buffer(reserve=10)
        self.assertEqual(buf.size, 10)
        self.assertEqual(len(buf.read()), 10)

        buf.orphan(20)
        self.assertEqual(buf.size, 20)
        self.assertEqual(len(buf.read()), 20)

        buf.orphan(100)
        self.assertEqual(buf.size, 100)
        self.assertEqual(len(buf.read()), 100)

        buf.orphan(-100)
        self.assertEqual(buf.size, 100)
        self.assertEqual(len(buf.read()), 100)


if __name__ == '__main__':
    unittest.main()



import moderngl_window
from moderngl_window import geometry
from pyrr import Matrix44


class PolygonOffset(moderngl_window.WindowConfig):
    title = "Polygon Offset"
    samples = 4

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.program_sphere = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 m_model;
            uniform mat4 m_proj;

            in vec3 in_position;
            in vec3 in_normal;

            out vec3 pos;
            out vec3 normal;

            void main() {
                vec4 p = m_model * vec4(in_position, 1.0);
                gl_Position =  m_proj * p;
                mat3 m_normal = inverse(transpose(mat3(m_model)));
                normal = m_normal * normalize(in_normal);
                pos = p.xyz;
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;
            uniform vec4 color = vec4(1.0, 0.0, 0.0, 1.0);

            in vec3 pos;
            in vec3 normal;

            void main() {
                float l = dot(normalize(-pos), normalize(normal));
                fragColor = color * (0.25 + abs(l) * 0.75);
            }
            """,
        )
        self.program_lines = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 m_model;
            uniform mat4 m_proj;

            in vec3 in_position;

            void main() {
                gl_Position = m_proj * m_model * vec4(in_position, 1.0);
            }

            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;

            void main() {
                fragColor = vec4(0.7, 0.7, 0.7, 1.0);
            }
            """,
        )

        self.sphere = geometry.sphere(radius=2.0, sectors=32, rings=16)
        self.projection = Matrix44.perspective_projection(60, self.wnd.aspect_ratio, 1, 100, dtype="f4")
        self.poly_offset_enabled = True

    def render(self, time: float, frame_time: float):
        self.ctx.clear()
        self.ctx.enable(self.ctx.DEPTH_TEST | self.ctx.CULL_FACE)

        self.program_lines["m_proj"].write(self.projection)
        self.program_sphere["m_proj"].write(self.projection)

        trans = Matrix44.from_translation((0, 0, -4.5), dtype="f4")
        rot = Matrix44.from_eulers((time/2, time/12.33, time/11.94), dtype="f4")
        matrix = rot @ trans
        self.program_lines["m_model"].write(matrix)
        self.program_sphere["m_model"].write(matrix)

        self.sphere.render(self.program_sphere)
        if self.poly_offset_enabled:
            self.ctx.polygon_offset = -1, -1
        self.ctx.wireframe = True
        self.sphere.render(self.program_lines)
        if self.poly_offset_enabled:
            self.ctx.polygon_offset = 0, 0
        self.ctx.wireframe = False

    def key_event(self, key, action, modifiers):
        keys = self.wnd.keys
        if action == keys.ACTION_PRESS and key == keys.SPACE:
            self.poly_offset_enabled = not self.poly_offset_enabled


PolygonOffset.run()



"""This example is a port to ModernGL of code by Nicolas P. Rougier from his "Python & OpenGL
for Scientific Visualization" free online book. Available under the (new) BSD License.

Book is available here:
https://github.com/rougier/python-opengl

Background information on this code:
https://github.com/rougier/python-opengl/blob/master/09-lines.rst

Original code on which this example is based:
https://github.com/rougier/python-opengl/blob/master/code/chapter-09/geom-path.py
"""

import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


# prepare geometry
def star(inner=0.45, outer=1.0, n=5):
    R = np.array([inner, outer] * n)
    T = np.linspace(-0.5 * np.pi, 1.5 * np.pi, 2 * n, endpoint=False)
    P = np.zeros((2 * n, 2))
    P[:, 0] = R * np.cos(T)
    P[:, 1] = R * np.sin(T)
    return np.vstack([P, P[0]])


def rect(x, y, w, h):
    return np.array([(x, y), (x + w, y), (x + w, y + h), (x, y + h), (x, y)])


def build_buffers(lines):
    """Prepare the buffers for multi-polyline rendering. Closed polyline must have their
    last point identical to their first point."""

    lines = [np.array(line, dtype="f4") for line in lines]

    indices = []
    reset_index = [-1]
    start_index = 0
    for line in lines:
        if np.all(line[0] == line[-1]):  # closed path
            idx = np.arange(len(line) + 3) - 1
            idx[0], idx[-2], idx[-1] = len(line) - 1, 0, 1
        else:
            idx = np.arange(len(line) + 2) - 1
            idx[0], idx[-1] = 0, len(line) - 1

        indices.append(idx + start_index)
        start_index += len(line)
        indices.append(reset_index)

    return np.vstack(lines).astype("f4"), np.concatenate(indices).astype("i4")


class RichLines(Example):
    title = "Rich Lines"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.line_prog = self.load_program("rich_lines.glsl")

        lines = [
            star(n=5) * 300 + (400, 400),
            star(n=8) * 150 + (900, 200),
            rect(900, 600, 150, 50),
            [(1200, 100), (1400, 200), (1300, 100)],
        ]

        vertex, index = build_buffers(lines)

        vbo = self.ctx.buffer(vertex)
        ibo = self.ctx.buffer(index)
        self.vao = self.ctx.simple_vertex_array(self.line_prog, vbo, "position",
                                                index_buffer=ibo)

        # Set the desired properties for the lines.
        # Note:
        # - round cap/ends are used if miter_limit < 0
        # - antialias value is in model space and should probably be scaled to be ~1.5px in
        #   screen space

        self.line_prog["linewidth"].value = 15
        self.line_prog["antialias"].value = 1.5
        self.line_prog["miter_limit"].value = -1
        self.line_prog["color"].value = 0, 0, 1, 1
        self.line_prog["projection"].write(
            Matrix44.orthogonal_projection(0, 1600, 800, 0, 0.5, -0.5, dtype="f4")
        )

    def render(self, time, frame_time):
        self.ctx.clear(1, 1, 1, 1)
        self.ctx.enable(moderngl.BLEND)
        self.vao.render(moderngl.LINE_STRIP_ADJACENCY)


if __name__ == '__main__':
    RichLines.run()



# This example opens an image, and offsets the red, green, and blue channels to create a glitchy RGB split effect.
from pathlib import Path
from array import array

from PIL import Image

import moderngl
import moderngl_window


class ImageProcessing(moderngl_window.WindowConfig):
    window_size = 3840 // 2, 2160 // 2
    resource_dir = Path(__file__).parent.resolve()
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.image_processing = ImageTransformer(self.ctx, self.window_size)
        self.texture = self.load_texture_2d("data/ball.png")

    def render(self, time, frame_time):
        self.image_processing.render(self.texture, target=self.ctx.screen)

        # Headless
        self.image_processing.render(self.texture)
        self.image_processing.write("output.png")


class ImageTransformer:

    def __init__(self, ctx, size, program=None):
        self.ctx = ctx
        self.size = size
        self.program = None
        self.fbo = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture(self.size, 4)]
        )

        # Create some default program if needed
        if not program:
            self.program = self.ctx.program(
                vertex_shader="""
                    #version 330

                    in vec2 in_position;
                    in vec2 in_uv;
                    out vec2 uv;

                    void main() {
                        gl_Position = vec4(in_position, 0.0, 1.0);
                        uv = in_uv;
                    }
                """,
                fragment_shader = """
                    #version 330

                    uniform sampler2D image;
                    in vec2 uv;
                    out vec4 out_color;

                    void main() {
                        // Get the Red, green, blue values from the image
                        float red = texture(image, uv).r;
                        // Offset green and blue
                        float green = texture(image, uv+(1.0/20.0)).g;
                        float blue = texture(image, uv+(2.0/20.0)).b;
                        float alpha = texture(image, uv).a;
                        
                        out_color = vec4(red, green, blue, alpha);
                    }
                """,          
            )

        # Fullscreen quad in NDC
        self.vertices = self.ctx.buffer(
            array(
                'f',
                [
                    # Triangle strip creating a fullscreen quad
                    # x, y, u, v
                    -1,  1, 0, 1,  # upper left
                    -1, -1, 0, 0, # lower left
                     1,  1, 1, 1, # upper right
                     1, -1, 1, 0, # lower right
                ]
            )
        )
        self.quad = self.ctx.vertex_array(
            self.program,
            [
                (self.vertices, '2f 2f', 'in_position', 'in_uv'),
            ]
        )

    def render(self, texture, target=None):
        if target:
            target.use()
        else:
            self.fbo.use()

        texture.use(0)
        self.quad.render(mode=moderngl.TRIANGLE_STRIP)

    def write(self, name):
        image = Image.frombytes("RGBA", self.fbo.size, self.fbo.read(components=4))
        image = image.transpose(Image.FLIP_TOP_BOTTOM)
        image.save(name, format="png")


if __name__ == "__main__":
    ImageProcessing.run()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_texture(self):
        tex = self.ctx.texture((4, 4), 4)
        tex.release()

    def test_texture_array(self):
        tex = self.ctx.texture_array((4, 4, 4), 4)
        tex.release()

    def test_texture_cube(self):
        tex = self.ctx.texture_cube((4, 4), 4)
        tex.release()

    def test_texture_3d(self):
        tex = self.ctx.texture3d((4, 4, 4), 4)
        tex.release()

    def test_renderbuffer(self):
        rbo = self.ctx.renderbuffer((4, 4))
        rbo.release()

    def test_buffer(self):
        buf = self.ctx.buffer(reserve=4)
        buf.release()

    def test_framebuffer(self):
        fbo = self.ctx.framebuffer(self.ctx.renderbuffer((4, 4)))
        fbo.release()

    def test_vertex_array(self):
        vao = self.ctx.vertex_array(self.ctx.program(
            vertex_shader="""
                #version 330
                void main() {
                    gl_Position = vec4(0.0, 0.0, 0.0, 1.0);
                }
            """,
        ), [])
        vao.release()

    def test_sampler(self):
        smp = self.ctx.sampler()
        smp.release()

    def test_scope(self):
        fbo = self.ctx.simple_framebuffer((4, 4))
        scope = self.ctx.scope(framebuffer=fbo)
        fbo.release()
        scope.release()

    # def test_query(self):
    #     qry = self.ctx.query()
    # NOTE: Query objects are not released



import struct
import platform
from array import array
import unittest
import numpy as np

import moderngl

from common import get_context


class TestCase(unittest.TestCase):
    """Test all combinations of transforms"""

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_transform(self):

        program = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec4 vert;
                out float vert_length;

                void main() {
                    vert_length = length(vert);
                }
            ''',
            varyings=['vert_length']
        )

        self.assertIs(program.geometry_input, None)
        self.assertIs(program.geometry_output, None)
        self.assertIs(program.is_transform, True)

        self.assertIn('vert', program)
        self.assertIn('vert_length', program)

    def test_vertex_points(self):
        """Transform with points"""
        data = (0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0)
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos;
            }
            """,
            varyings=["out_pos"],
        )

        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        self.assertIs(vao.mode, moderngl.POINTS)
        buffer2 = self.ctx.buffer(reserve=buffer.size)
        vao.transform(buffer2, mode=self.ctx.POINTS)
        self.assertEqual(data, struct.unpack(f"{len(data)}f", buffer.read()))

    def test_vertex_lines(self):
        data = (
            0.0, 0.0,
            1.0, 1.0,
            2.0, 2.0,
            3.0, 3.0,
        )
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos * 2;
            }
            """,
            varyings=["out_pos"],
        )

        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        buffer2 = self.ctx.buffer(reserve=buffer.size)
        vao.transform(buffer2, mode=self.ctx.LINES)
        self.assertEqual(tuple(v * 2 for v in data), struct.unpack(f"{len(data)}f", buffer2.read()))

    # NOTE: This fails in tests and may be non-standard or require higher gl version
    # def test_vertex_lines_adjacency(self):
    #     if platform.system().lower() in ["darwin"]:
    #         self.skipTest('Tranforms with adjacency primitives now working on OSX')

    #     data = (
    #         1.0, 1.0,
    #         2.0, 2.0,
    #         3.0, 3.0,
    #         4.0, 4.0,
    #     )
    #     buffer = self.ctx.buffer(array('f', data))
    #     program = self.ctx.program(
    #         vertex_shader="""
    #         #version 330

    #         in vec2 in_pos;
    #         out vec2 out_pos;

    #         void main() {
    #             out_pos = in_pos;
    #         }
    #         """,
    #         varyings=["out_pos"],
    #     )
    #     vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
    #     buffer2 = self.ctx.buffer(reserve=4 * 4)
    #     vao.transform(buffer2, mode=self.ctx.LINES_ADJACENCY)

    #     # Start end end line is removed
    #     self.assertEqual(tuple(data[2:6]), struct.unpack(f"4f", buffer2.read()))

    def test_vertex_line_strip(self):
        data = (
            1.0, 1.0,
            2.0, 2.0,
            3.0, 3.0,
            4.0, 4.0,
        )
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos;
            }
            """,
            varyings=["out_pos"],
        )
        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        buffer2 = self.ctx.buffer(reserve=4 * 12)
        vao.transform(buffer2, mode=self.ctx.LINE_STRIP)
        self.assertEqual(
            (1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0),
            struct.unpack("12f", buffer2.read())
        )

    # NOTE: Fails in tests and might be non-standard or require higher gl version
    # def test_vertex_line_strip_adjacency(self):
    #     if platform.system().lower() in ["darwin"]:
    #         self.skipTest('Tranforms with adjacency primitives now working on OSX')

    #     data = (
    #         1.0, 1.0,
    #         2.0, 2.0,
    #         3.0, 3.0,
    #         4.0, 4.0,
    #         5.0, 5.0,
    #     )
    #     buffer = self.ctx.buffer(array('f', data))
    #     program = self.ctx.program(
    #         vertex_shader="""
    #         #version 330

    #         in vec2 in_pos;
    #         out vec2 out_pos;

    #         void main() {
    #             out_pos = in_pos;
    #         }
    #         """,
    #         varyings=["out_pos"],
    #     )
    #     vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
    #     buffer2 = self.ctx.buffer(reserve=4 * 8)
    #     vao.transform(buffer2, mode=self.ctx.LINE_STRIP_ADJACENCY)
    #     self.assertEqual(
    #         (2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0),
    #         struct.unpack("8f", buffer2.read())
    #     )

    def test_vertex_triangles(self):
        data = (
            1.0, 1.0,
            2.0, 2.0,
            3.0, 3.0,
            4.0, 4.0,
            5.0, 5.0,
            6.0, 6.0,
        )
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos * 2;
            }
            """,
            varyings=["out_pos"],
        )
        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        buffer2 = self.ctx.buffer(reserve=4 * 12)
        vao.transform(buffer2, mode=self.ctx.TRIANGLES)
        self.assertEqual(
            tuple(v * 2 for v in data),
            struct.unpack("12f", buffer2.read())
        )

    def test_vertex_triangles_adjacency(self):
        if platform.system().lower() in ["darwin"]:
            self.skipTest('Tranforms with adjacency primitives now working on OSX')
        data = (
            1.0, 1.0, # outer points
            2.0, 2.0,
            3.0, 3.0,
            4.0, 4.0,
            5.0, 5.0,
            6.0, 6.0,
            7.0, 7.0, # inner triangle
            8.0, 8.0,
            9.0, 9.0,
        )
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos;
            }
            """,
            varyings=["out_pos"],
        )
        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        buffer2 = self.ctx.buffer(reserve=4 * 6)
        vao.transform(buffer2, mode=self.ctx.TRIANGLES_ADJACENCY)
        self.assertEqual(
            (1.0, 1.0, 3.0, 3.0, 5.0, 5.0),
            struct.unpack("6f", buffer2.read())
        )

    def test_vertex_triangle_strip(self):

        data = (
            1.0, 1.0,
            2.0, 2.0,
            3.0, 3.0,
            4.0, 4.0,
        )
        buffer = self.ctx.buffer(array('f', data))
        program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_pos;
            out vec2 out_pos;

            void main() {
                out_pos = in_pos;
            }
            """,
            varyings=["out_pos"],
        )
        vao = self.ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
        buffer2 = self.ctx.buffer(reserve=4 * 12)
        vao.transform(buffer2, mode=self.ctx.TRIANGLE_STRIP)
        self.assertEqual(
            (1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 2.0, 2.0, 4.0, 4.0),
            struct.unpack("12f", buffer2.read())
        )        

    def test_vertex_triangle_fan(self):
        vertices = np.array(
            [
                0.0, 0.0,  # Initial triangle
                1.0, 0.0,
                1.0, 1.0,
                0.0, 1.0,  # second triangle 
            ],
            dtype='f4',
        )

        prog = self.ctx.program(vertex_shader=
            """
            #version 330
            in vec2 in_vert;
            out vec2 out_vert;
            void main(void) {
                out_vert = in_vert;
            }
            """,
            varyings=['out_vert'],
        )

        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, '2f', 'in_vert')])
        N = 6  # 6 vertices in end buffer
        buffer = self.ctx.buffer(reserve=N * 4 * 2)
        vao.transform(buffer, vertices=4, mode=self.ctx.TRIANGLE_FAN)

        data = struct.unpack('{}f'.format(N * 2), buffer.read())
        self.assertEqual(data,
            (
                0.0, 0.0,  # Initial triangle
                1.0, 0.0,
                1.0, 1.0,
                0.0, 0.0,  # second triangle 
                1.0, 1.0, 
                0.0, 1.0,
            )
        )

    # ------ Geometry shader -------

    def test_geometry_points(self):
        data = (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(points) in;
            layout(points, max_vertices=1) out;

            void main() {
                out_value = gl_in[0].gl_Position.x * 2;
                EmitVertex();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=vbo.size)
        vao.transform(buffer, mode=moderngl.POINTS)
        self.assertEqual(tuple(v * 2 for v in data), struct.unpack('6f', buffer.read()))

    def test_geometry_lines(self):
        data = (
            1.0, 2.0,
            3.0, 4.0,
            5.0, 6.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(lines) in;
            layout(line_strip, max_vertices=2) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=vbo.size)
        vao.transform(buffer, mode=moderngl.LINES)
        self.assertEqual(data, struct.unpack('6f', buffer.read()))

    def test_geometry_line_loop(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(lines) in;
            layout(line_strip, max_vertices=2) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 12)
        vao.transform(buffer, mode=moderngl.LINE_LOOP)
        self.assertEqual(
            (1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0, 5.0, 6.0, 6.0, 1.0),
            struct.unpack('12f', buffer.read())
        )

    def test_geometry_line_strip(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(lines) in;
            layout(line_strip, max_vertices=2) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 8)
        vao.transform(buffer, mode=moderngl.LINE_LOOP)
        self.assertEqual(
            (1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0),
            struct.unpack('8f', buffer.read())
        )

    def test_geometry_lines_adjacency(self):
        data = (
            1.0, 2.0, 3.0, 4.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(lines_adjacency) in;
            layout(line_strip, max_vertices=2) out;

            void main() {
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 2)
        vao.transform(buffer, mode=moderngl.LINES_ADJACENCY)
        self.assertEqual(
            (2.0, 3.0),
            struct.unpack('2f', buffer.read())
        )

    def test_geometry_line_strip_adjacency(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(lines_adjacency) in;
            layout(line_strip, max_vertices=2) out;

            void main() {
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 4)
        vao.transform(buffer, mode=moderngl.LINE_STRIP_ADJACENCY)
        self.assertEqual(
            (2.0, 3.0, 3.0, 4.0),
            struct.unpack('4f', buffer.read())
        )

    # ---

    def test_geometry_triangles(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(triangles) in;
            layout(triangle_strip, max_vertices=3) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=vbo.size)
        vao.transform(buffer, mode=moderngl.TRIANGLES)
        self.assertEqual(data, struct.unpack('6f', buffer.read()))

    def test_geometry_triangle_strip(self):
        data = (
            1.0, 2.0, 3.0, 4.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(triangles) in;
            layout(triangle_strip, max_vertices=3) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 6)
        vao.transform(buffer, mode=moderngl.TRIANGLE_STRIP)
        self.assertEqual(
            (1.0, 2.0, 3.0, 3.0, 2.0, 4.0),
            struct.unpack('6f', buffer.read())
        )

    def test_geometry_triangle_fan(self):
        data = (
            0.0, 1.0, 2.0, 3.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(triangles) in;
            layout(triangle_strip, max_vertices=3) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[1].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 6)
        vao.transform(buffer, mode=moderngl.TRIANGLE_FAN)
        self.assertEqual(
            (0.0, 1.0, 2.0, 0.0, 2.0, 3.0),
            struct.unpack('6f', buffer.read())
        )

    def test_geometry_triangles_adjacency(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(triangles_adjacency) in;
            layout(triangle_strip, max_vertices=3) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                out_value = gl_in[4].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 3)
        vao.transform(buffer, mode=moderngl.TRIANGLES_ADJACENCY)
        self.assertEqual(
            (1.0, 3.0, 5.0),
            struct.unpack('3f', buffer.read())
        )

    def test_geometry_triangle_strip_adjacency(self):
        data = (
            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0,
        )
        vertices = np.array(data, dtype='f4')
        prog = self.ctx.program(
            vertex_shader=
            """
            #version 330

            in float in_value;

            void main() {
                gl_Position = vec4(in_value);
            }
            """,
            geometry_shader="""
            #version 330

            out float out_value;

            layout(triangles_adjacency) in;
            layout(triangle_strip, max_vertices=3) out;

            void main() {
                out_value = gl_in[0].gl_Position.x;
                EmitVertex();
                out_value = gl_in[2].gl_Position.x;
                EmitVertex();
                out_value = gl_in[4].gl_Position.x;
                EmitVertex();
                EndPrimitive();
            }
            """,
            varyings=['out_value'],
        )
        vbo = self.ctx.buffer(vertices)
        vao = self.ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
        buffer = self.ctx.buffer(reserve=4 * 6)
        vao.transform(buffer, mode=moderngl.TRIANGLE_STRIP_ADJACENCY)
        self.assertEqual(
            (1.0, 3.0, 5.0, 5.0, 3.0, 7.0),
            struct.unpack('6f', buffer.read())
        )

    def gl_error(self, raise_exception=True):
        error = self.ctx.error
        if error != "GL_NO_ERROR" and raise_exception:
            raise ValueError(error)


if __name__ == '__main__':
    unittest.main()



import moderngl
from basic_colors_and_texture import ColorsAndTexture


class RenderToTexture(ColorsAndTexture):
    title = "Render to Texture"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.texture1 = self.texture
        self.texture2 = self.ctx.texture(self.wnd.size, 3)
        depth_attachment = self.ctx.depth_renderbuffer(self.wnd.size)
        self.fbo = self.ctx.framebuffer(self.texture2, depth_attachment)

    def render(self, time, frame_time):
        for mode in ['render_to_texture', 'render_to_window']:
            if mode == 'render_to_texture':
                self.texture = self.texture1
                self.fbo.clear(1.0, 1.0, 1.0)
                self.fbo.use()
            else:
                self.texture = self.texture2
                self.ctx.screen.use()
            super().render(time, frame_time)


if __name__ == '__main__':
    RenderToTexture.run()



'''
    Renders a floating, oscillating, 3d islans with lights
'''
import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


class ColorsAndTexture(Example):
    gl_version = (3, 3)
    title = "Colors and Textures"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_position;
                in vec3 in_normal;
                in vec2 in_texcoord_0;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_text;

                void main() {
                    gl_Position = Mvp * vec4(in_position, 1.0);
                    v_vert = in_position;
                    v_norm = in_normal;
                    v_text = in_texcoord_0;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 Light;
                uniform vec3 Color;
                uniform bool UseTexture;
                uniform sampler2D Texture;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    float lum = clamp(dot(normalize(Light - v_vert), normalize(v_norm)), 0.0, 1.0) * 0.8 + 0.2;
                    if (UseTexture) {
                        f_color = vec4(texture(Texture, v_text).rgb * lum, 1.0);
                    } else {
                        f_color = vec4(Color * lum, 1.0);
                    }
                }
            ''',
        )

        self.mvp = self.prog['Mvp']
        self.light = self.prog['Light']
        self.color = self.prog['Color']
        self.use_texture = self.prog['UseTexture']

        # Note: This is a fairly manual way to loading and rendering wavefront files.
        # There are easier ways when loading multiple objects in a single obj file.

        # Load obj files
        self.scene_ground = self.load_scene('scene-1-ground.obj')
        self.scene_grass = self.load_scene('scene-1-grass.obj')
        self.scene_billboard = self.load_scene('scene-1-billboard.obj')
        self.scene_holder = self.load_scene('scene-1-billboard-holder.obj')
        self.scene_image = self.load_scene('scene-1-billboard-image.obj')

        # Extract the VAOs from the scene
        self.vao_ground = self.scene_ground.root_nodes[0].mesh.vao.instance(self.prog)
        self.vao_grass = self.scene_grass.root_nodes[0].mesh.vao.instance(self.prog)
        self.vao_billboard = self.scene_billboard.root_nodes[0].mesh.vao.instance(self.prog)
        self.vao_holder = self.scene_holder.root_nodes[0].mesh.vao.instance(self.prog)
        self.vao_image = self.scene_image.root_nodes[0].mesh.vao.instance(self.prog)

        # texture on billboard
        self.texture = self.load_texture_2d('infographic-1.jpg')

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (47.697, -8.147, 24.498),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 1.0),
        )

        rotate = Matrix44.from_z_rotation(np.sin(time) * 0.5 + 0.2)

        self.use_texture.value = False

        self.light.value = (67.69, -8.14, 52.49)
        self.mvp.write((proj * lookat * rotate).astype('f4'))

        self.color.value = (0.67, 0.49, 0.29)
        self.vao_ground.render()

        self.color.value = (0.46, 0.67, 0.29)
        self.vao_grass.render()

        self.color.value = (1.0, 1.0, 1.0)
        self.vao_billboard.render()

        self.color.value = (0.2, 0.2, 0.2)
        self.vao_holder.render()

        self.use_texture.value = True
        self.texture.use()

        self.vao_image.render()


if __name__ == '__main__':
    ColorsAndTexture.run()



"""
Using pycairo with moderngl.

We simply create a screen aligned quad with texture coordinates
to render the uploaded texture from cairo.

Textures in OpenGL are stored "upside-down" so we build
a vertex array with inverted y coordinates

"""
import math
from array import array

import cairo
import moderngl
from moderngl_window import geometry
from ported._example import Example


class CairoExample(Example):
    title = "Cairo Integration"
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.texture = self.render_cairo_to_texture(512, 512)
        self.prog = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec3 in_position;
            in vec2 in_texcoord_0;
            out vec2 uv;
            void main() {
                gl_Position = vec4(in_position, 1.0);
                uv = in_texcoord_0;
            }
            """,
            fragment_shader="""
            #version 330
            uniform sampler2D texture0;
            in vec2 uv;
            out vec4 outColor;
            void main() {
                outColor = texture(texture0, uv);
            }
            """,
        )
        # Create a simple screen rectangle. The texture coordinates
        # are reverted on the y axis here to make the cairo texture appear correctly.
        vertices = [
            # x, y | u, v
            -1,  1,  0, 0,
            -1, -1,  0, 1,
             1,  1,  1, 0,
             1, -1,  1, 1,
        ]
        self.screen_rectangle = self.ctx.vertex_array(
            self.prog,
            [
                (
                    self.ctx.buffer(array('f', vertices)),
                    '2f 2f',
                    'in_position', 'in_texcoord_0',
                )
            ],
        )

    def render(self, time, frame_time):
        self.texture.use(location=0)
        self.screen_rectangle.render(mode=moderngl.TRIANGLE_STRIP)

    def render_cairo_to_texture(self, width, height):
        # Draw with cairo to surface
        x, y, radius = (250, 250, 200)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        ctx = cairo.Context(surface)
        ctx.set_line_width(15)
        ctx.arc(x, y, radius, 0, 2.0 * math.pi)
        ctx.set_source_rgb(0.8, 0.8, 0.8)
        ctx.fill_preserve()
        ctx.set_source_rgb(1, 1, 0)
        ctx.stroke()

        ctx.move_to(20, 30)
        ctx.select_font_face("Purisa", cairo.FONT_SLANT_NORMAL,
                            cairo.FONT_WEIGHT_NORMAL)
        ctx.set_font_size(13)
        ctx.set_source_rgb(0.8, 0.8, 0.8)
        ctx.show_text("Example Text")
        # Copy surface to texture
        texture = self.ctx.texture((width, height), 4, data=surface.get_data())
        texture.swizzle = 'BGRA' # use Cairo channel order (alternatively, the shader could do the swizzle)
        return texture

if __name__ == "__main__":
    CairoExample.run()



from typing import Any, Tuple

__all__ = ['Query']


class Query:
    """This class represents a Query object."""

    __slots__ = ['mglo', 'crender', 'ctx', 'extra']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self.crender = None  #: ConditionalRender: Can be used in a ``with`` statement.
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self) -> str:
        return '<Query>'

    def __hash__(self) -> int:
        return id(self)

    def __enter__(self):
        self.mglo.begin()
        return self

    def __exit__(self, *args: Tuple[Any]):
        self.mglo.end()

    @property
    def samples(self) -> int:
        """int: The number of samples passed."""
        return self.mglo.samples

    @property
    def primitives(self) -> int:
        """int: The number of primitives generated."""
        return self.mglo.primitives

    @property
    def elapsed(self) -> int:
        """int: The time elapsed in nanoseconds."""
        return self.mglo.elapsed



"""
This modules is really unisable for ref counting. Just an experiment!
"""
import gc
import sys
import ctypes

def print_referrers(obj):
    refs = gc.get_referrers(obj)
    for ref in refs:
        if is_module(ref):
            print("-->", is_module(ref))
        else:
            print("--> ", type(ref), ref)


def is_module(obj):
    if isinstance(obj, dict) and obj.get("__name__"):
        return obj.get("__name__")
    return None

def ref_c_1(obj):
    """+ 1"""
    return ctypes.c_long.from_address(id(obj)).value

def ref_c_2(obj):
    """+ 2"""
    return sys.getrefcount(obj)



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;
#         in vec2 pos;
#         in float scale;
#         in vec3 color;

#         out vec3 v_color;

#         void main() {
#             v_color = color;
#             gl_Position = vec4(pos + vert * scale, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# # Vertex coordinates stored in vbo1
# #
# #     B------D
# #     |      |
# #     A------C

# vbo1 = ctx.buffer(struct.pack(
#     '8f',
#     -0.5, -0.5,
#     -0.5, 0.5,
#     0.5, -0.5,
#     0.5, 0.5,
# ))

# # Vertex colors stored in vbo2
# #
# #     A, B are green
# #     C, D are blue

# vbo2 = ctx.buffer(struct.pack(
#     '12f',
#     0.0, 1.0, 0.0,
#     0.0, 1.0, 0.0,
#     0.0, 0.0, 1.0,
#     0.0, 0.0, 1.0,
# ))

# # (Per instance) positions and scales stored in vbo3
# # There are 8 (position, scale) pairs

# vbo3 = ctx.buffer(struct.pack(
#     '24f',
#     0.5, 0.0, 0.3,
#     0.35, 0.35, 0.2,
#     0.0, 0.5, 0.3,
#     -0.35, 0.35, 0.2,
#     -0.5, 0.0, 0.3,
#     -0.35, -0.35, 0.2,
#     0.0, -0.5, 0.3,
#     0.35, -0.35, 0.2,
# ))

# # Index buffer (also called element buffer)
# # There are 2 trianges to render
# #
# #     A, B, C
# #     B, C, D

# ibo = ctx.buffer(struct.pack('6i', 0, 1, 2, 1, 2, 3))

# # The vao_content is a list of 3-tuples (buffer, format, attribs)
# # the format can have an empty or '/v', '/i', '/r' ending.
# # '/v' attributes are the default
# # '/i` attributes are per instance attributes
# # '/r' attributes are default values for the attributes (per render attributes)

# vao_content = [
#     (vbo1, '2f', ['vert']),
#     (vbo2, '3f', ['color']),
#     (vbo3, '2f1f/i', ['pos', 'scale']),
# ]

# vao = ctx.vertex_array(prog, vao_content, ibo)

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render(instances=8)



import moderngl


class ContextManager:
    ctx = None

    @staticmethod
    def get_default_context(allow_fallback_standalone_context=True) -> moderngl.Context:
        '''
            Default context
        '''

        if ContextManager.ctx is None:
            try:
                ContextManager.ctx = moderngl.create_context()
            except:
                if allow_fallback_standalone_context:
                    ContextManager.ctx = moderngl.create_standalone_context()
                else:
                    raise

        return ContextManager.ctx



import os

import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


def terrain(size):
    vertices = np.dstack(np.mgrid[0:size, 0:size][::-1]) / size
    temp = np.dstack([np.arange(0, size * size - size), np.arange(size, size * size)])
    index = np.pad(temp.reshape(size - 1, 2 * size), [[0, 0], [0, 1]], 'constant', constant_values=-1)
    return vertices, index


class WireframeTerrain(Example):
    title = "Wireframe Terrain"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;
                uniform sampler2D Heightmap;

                in vec2 in_vert;
                in vec3 in_color;

                out vec3 v_color;

                void main() {
                    v_color = in_color;
                    float height = texture(Heightmap, in_vert.xy).r * 0.5;
                    gl_Position = Mvp * vec4(in_vert.xy - 0.5, height, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                in vec3 v_color;
                out vec4 f_color;

                void main() {
                    f_color = vec4(v_color, 1.0);
                }
            ''',
        )

        self.mvp = self.prog['Mvp']

        vertices, index = terrain(32)

        self.vbo = self.ctx.buffer(vertices.astype('f4'))
        self.ibo = self.ctx.buffer(index.astype('i4'))

        vao_content = [
            (self.vbo, '2f', 'in_vert'),
        ]

        self.vao = self.ctx.vertex_array(self.prog, vao_content, self.ibo)
        self.texture = self.load_texture_2d('noise.jpg')

    def render(self, time, frame_time):
        angle = time * 0.2

        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)
        self.ctx.wireframe = True

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (np.cos(angle), np.sin(angle), 0.8),
            (0.0, 0.0, 0.1),
            (0.0, 0.0, 1.0),
        )

        self.texture.use()
        self.mvp.write((proj * lookat).astype('f4'))
        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    WireframeTerrain.run()



from typing import Any, Optional, Tuple, Union

from _moderngl import InvalidObject

from .buffer import Buffer

__all__ = ['TextureArray']


class TextureArray:
    """
    An Array Texture is a Texture where each mipmap level contains an array of images of the same size.

    Array textures may have Mipmaps, but each mipmap
    in the texture has the same number of levels.

    A TextureArray object cannot be instantiated directly, it requires a context.
    Use :py:meth:`Context.texture_array` to create one.
    """

    __slots__ = ['mglo', '_size', '_components', '_samples', '_dtype', '_depth', '_glo', 'ctx', 'extra']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._size = (None, None, None)
        self._components = None
        self._samples = None
        self._dtype = None
        self._depth = None
        self._glo = None
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self):
        if hasattr(self, '_glo'):
            return f"<{self.__class__.__name__}: {self._glo}>"
        else:
            return f"<{self.__class__.__name__}: INCOMPLETE>"

    def __eq__(self, other: Any):
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def repeat_x(self) -> bool:
        """
        bool: The x repeat flag for the texture (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            texture.repeat_x = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            texture.repeat_x = False
        """
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value: bool) -> None:
        self.mglo.repeat_x = value

    @property
    def repeat_y(self) -> bool:
        """
        bool: The y repeat flag for the texture (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            texture.repeat_y = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            texture.repeat_y = False
        """
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value: bool) -> None:
        self.mglo.repeat_y = value

    @property
    def filter(self) -> Tuple[int, int]:
        """
        tuple: The minification and magnification filter for the texture.

        (Default ``(moderngl.LINEAR. moderngl.LINEAR)``)

        Example::

            texture.filter == (moderngl.NEAREST, moderngl.NEAREST)
            texture.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)
            texture.filter == (moderngl.NEAREST_MIPMAP_LINEAR, moderngl.NEAREST)
            texture.filter == (moderngl.LINEAR_MIPMAP_NEAREST, moderngl.NEAREST)
        """
        return self.mglo.filter

    @filter.setter
    def filter(self, value: Tuple[int, int]) -> None:
        self.mglo.filter = value

    @property
    def swizzle(self) -> str:
        """
        str: The swizzle mask of the texture (Default ``'RGBA'``).

        The swizzle mask change/reorder the ``vec4`` value returned by the ``texture()`` function
        in a GLSL shaders. This is represented by a 4 character string were each
        character can be::

            'R' GL_RED
            'G' GL_GREEN
            'B' GL_BLUE
            'A' GL_ALPHA
            '0' GL_ZERO
            '1' GL_ONE

        Example::

            # Alpha channel will always return 1.0
            texture.swizzle = 'RGB1'

            # Only return the red component. The rest is masked to 0.0
            texture.swizzle = 'R000'

            # Reverse the components
            texture.swizzle = 'ABGR'
        """
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value: str) -> None:
        self.mglo.swizzle = value

    @property
    def anisotropy(self) -> float:
        """
        float: Number of samples for anisotropic filtering (Default ``1.0``).

        The value will be clamped in range ``1.0`` and ``ctx.max_anisotropy``.

        Any value greater than 1.0 counts as a use of anisotropic filtering::

            # Disable anisotropic filtering
            texture.anisotropy = 1.0

            # Enable anisotropic filtering suggesting 16 samples as a maximum
            texture.anisotropy = 16.0
        """
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value: float) -> None:
        self.mglo.anisotropy = value

    @property
    def width(self) -> int:
        """int: The width of the texture array."""
        return self._size[0]

    @property
    def height(self) -> int:
        """int: The height of the texture array."""
        return self._size[1]

    @property
    def layers(self) -> int:
        """int: The number of layers of the texture array."""
        return self._size[2]

    @property
    def size(self) -> tuple:
        """tuple: The size of the texture array."""
        return self._size

    @property
    def components(self) -> int:
        """int: The number of components of the texture array."""
        return self._components

    @property
    def dtype(self) -> str:
        """str: Data type."""
        return self._dtype

    @property
    def glo(self) -> int:
        """
        int: The internal OpenGL object.

        This values is provided for debug purposes only.
        """
        return self._glo

    def read(self, *, alignment: int = 1) -> bytes:
        """
        Read the pixel data as bytes into system memory.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.

        Returns:
            bytes
        """
        return self.mglo.read(alignment)

    def read_into(
        self,
        buffer: Any,
        *,
        alignment: int = 1,
        write_offset: int = 0,
    ) -> None:
        """
        Read the content of the texture array into a bytearray or :py:class:`~moderngl.Buffer`.

        The advantage of reading into a :py:class:`~moderngl.Buffer` is that pixel data
        does not need to travel all the way to system memory::

            # Reading pixel data into a bytearray
            data = bytearray(8)
            texture = ctx.texture((2, 2, 2), 1)
            texture.read_into(data)

            # Reading pixel data into a buffer
            data = ctx.buffer(reserve=8)
            texture = ctx.texture((2, 2, 2), 1)
            texture.read_into(data)

        Args:
            buffer (Union[bytearray, Buffer]): The buffer that will receive the pixels.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
            write_offset (int): The write offset.
        """
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(
        self,
        data: Any,
        viewport: Optional[Union[Tuple[int, int, int], Tuple[int, int, int, int, int, int]]] = None,
        *,
        alignment: int = 1,
    ) -> None:
        r"""
        Update the content of the texture array from byte data or a moderngl :py:class:`~moderngl.Buffer`.

        The ``viewport`` can be used for finer control of where the
        data should be written in the array. The valid versions are::

            # Writing multiple layers from the begining of the texture
            texture.write(data, viewport=(width, hight, num_layers))

            # Writing sub-sections of the array
            texture.write(data, viewport=(x, y, layer, width, height, num_layers))

        Like with other texture types we can also use bytes or :py:class:`~moderngl.Buffer`
        as a source::

            # Using a moderngl buffer
            data = ctx.buffer(reserve=8)
            texture = ctx.texture_array((2, 2, 2), 1)
            texture.write(data)

            # Using byte data from system memory
            data = b"\xff\xff\xff\xff\xff\xff\xff\xff"
            texture = ctx.texture_array((2, 2, 2), 1)
            texture.write(data)

        Args:
            data (bytes): The pixel data.
            viewport (tuple): The viewport.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
        """
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base: int = 0, max_level: int = 1000) -> None:
        """
        Generate mipmaps.

        This also changes the texture filter to ``LINEAR_MIPMAP_LINEAR, LINEAR``
        (Will be removed in ``6.x``)

        Keyword Args:
            base (int): The base level
            max_level (int): The maximum levels to generate
        """
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location: int = 0) -> None:
        """
        Bind the texture to a texture unit.

        The location is the texture unit we want to bind the texture.
        This should correspond with the value of the ``sampler2DArray``
        uniform in the shader because samplers read from the texture
        unit we assign to them::

            # Define what texture unit our two sampler2DArray uniforms should represent
            program['texture_a'] = 0
            program['texture_b'] = 1
            # Bind textures to the texture units
            first_texture.use(location=0)
            second_texture.use(location=1)

        Args:
            location (int): The texture location/unit.
        """
        self.mglo.use(location)

    def bind_to_image(self, unit: int, read: bool = True, write: bool = True, level: int = 0, format: int = 0) -> None:
        """
        Bind a texture to an image unit (OpenGL 4.2 required).

        This is used to bind textures to image units for shaders.
        The idea with image load/store is that the user can bind
        one of the images in a Texture to a number of image binding points
        (which are separate from texture image units). Shaders can read
        information from these images and write information to them,
        in ways that they cannot with textures.

        It's important to specify the right access type for the image.
        This can be set with the ``read`` and ``write`` arguments.
        Allowed combinations are:

        - **Read-only**: ``read=True`` and ``write=False``
        - **Write-only**: ``read=False`` and ``write=True``
        - **Read-write**: ``read=True`` and ``write=True``

        ``format`` specifies the format that is to be used when performing
        formatted stores into the image from shaders. ``format`` must be
        compatible with the texture's internal format. **By default the format
        of the texture is passed in. The format parameter is only needed
        when overriding this behavior.**

        Note that we bind the texture array as layered to make
        all the layers accessible. This can be updated to map
        single layers in the future.

        More information:

        - https://www.khronos.org/opengl/wiki/Image_Load_Store
        - https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glBindImageTexture.xhtml

        Args:
            unit (int): Specifies the index of the image unit to which to bind the texture
            texture (:py:class:`moderngl.Texture`): The texture to bind
        Keyword Args:
            read (bool): Allows the shader to read the image (default: ``True``)
            write (bool): Allows the shader to write to the image (default: ``True``)
            level (int): Level of the texture to bind (default: ``0``).
            format (int): (optional) The OpenGL enum value representing the format (defaults to the texture's format)
        """
        self.mglo.bind(unit, read, write, level, format)

    def release(self) -> None:
        """Release the ModernGL object."""
        if self.mglo is not None:
            self.mglo.release()
            self.mglo = InvalidObject()



import numpy as np

import moderngl
from ported._example import Example


class Conway(Example):
    title = "Conway's Game of Life Labyrinth"
    window_size = (640, 640)
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        width, height = self.wnd.width, self.wnd.height
        canvas = np.array([0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0]).astype('f4')
        pixels = np.round(np.random.rand(width, height)).astype('f4')
        grid = np.dstack(np.mgrid[0:height, 0:width][::-1]).astype('i4')

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 v_text;

                void main() {
                    v_text = in_vert;
                    gl_Position = vec4(in_vert * 2.0 - 1.0, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;

                in vec2 v_text;
                out vec4 f_color;

                void main() {
                    f_color = texture(Texture, v_text);
                }
            ''',
        )

        self.transform = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform sampler2D Texture;
                uniform int Width;
                uniform int Height;

                in ivec2 in_text;
                out float out_vert;

                #define LIVING 0.0
                #define DEAD 1.0

                bool cell(int x, int y) {
                    return texelFetch(Texture, ivec2((x + Width) % Width, (y + Height) % Height), 0).r < 0.5;
                }

                void main() {
                    bool living = cell(in_text.x, in_text.y);

                    int neighbours = 0;
                    if (cell(in_text.x - 1, in_text.y - 1)) neighbours++;
                    if (cell(in_text.x - 1, in_text.y + 0)) neighbours++;
                    if (cell(in_text.x - 1, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y - 1)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y + 0)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 0, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 0, in_text.y - 1)) neighbours++;

                    if (living) {
                        out_vert = (neighbours >= 2 && neighbours <= 4) ? LIVING : DEAD;
                    } else {
                        out_vert = (neighbours == 3) ? LIVING : DEAD;
                    }
                }
            ''',
            varyings=['out_vert']
        )

        self.transform['Width'].value = width
        self.transform['Height'].value = height

        self.texture = self.ctx.texture((width, height), 1, pixels.tobytes(), dtype='f4')
        self.texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.texture.swizzle = 'RRR1'
        self.texture.use()

        self.vbo = self.ctx.buffer(canvas)
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

        self.text = self.ctx.buffer(grid)
        self.tao = self.ctx.simple_vertex_array(self.transform, self.text, 'in_text')
        self.pbo = self.ctx.buffer(reserve=pixels.nbytes)

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)

        self.tao.transform(self.pbo)
        self.texture.write(self.pbo)

        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    Conway.run()



import unittest

import moderngl
from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context(require=430)
        if not cls.ctx:
            raise unittest.SkipTest('Compute shader not supported')

    def test_compute_shader(self):
        if self.ctx.version_code < 430:
            self.skipTest('OpenGL 4.3 is not supported')

        src = '''
            #version 430
            layout (local_size_x = 1, local_size_y = 1, local_size_z = 1) in;

            struct Something
            {
                vec4 pos;
            };

            layout(std430, binding=0) buffer something_in
            {
                Something entries[];
            } In;

            // output into the buffer bound to 1
            layout(std430, binding=1) buffer something_out
            {
                Something entries[];
            } Out;

            uniform float multiplier;

            void main() {
                int x = int(gl_GlobalInvocationID);
                Something entry = In.entries[x];

                Out.entries[x] = Something(entry.pos * multiplier);
            }
        '''

        self.buffer_in = self.ctx.buffer(reserve=16)
        self.buffer_out = self.ctx.buffer(reserve=16)

        self.buffer_in.bind_to_storage_buffer(binding=0)
        self.buffer_out.bind_to_storage_buffer(binding=1)

        compute_shader = self.ctx.compute_shader(src)

        self.assertIsInstance(compute_shader.glo, int)
        self.assertIsNotNone(compute_shader.mglo)
        compute_shader.extra = {'name': 'test'}
        self.assertEqual(compute_shader.extra, {'name': 'test'})
        self.assertIsInstance(compute_shader.ctx, moderngl.Context)

        compute_shader['multiplier'] = 1.0
        compute_shader['multiplier'].value = 1.0
        compute_shader.run()
        self.assertIsNotNone(compute_shader.get('multiplier', None))
        self.assertEqual(compute_shader, compute_shader)
        self.assertEqual([i for i in compute_shader], ['multiplier'])
        compute_shader.release()


if __name__ == '__main__':
    unittest.main()



import moderngl

ctx = moderngl.create_standalone_context()

prog = ctx.program(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec3 in_color;

        out vec3 v_color;

        void main() {
            v_color = in_color;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330

        in vec3 v_color;

        out vec3 f_color;

        void main() {
            f_color = v_color;
        }
    ''',
)



import struct
from colorsys import hls_to_rgb as hls
from random import uniform

import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


def random_color():
    return hls(uniform(0.0, 1.0), 0.5, 0.5)


cars = []
cars += [{'color': random_color(),
          'pos': (1.5, i * 2.0 - 9.0, 0.0),
          'angle': uniform(-0.5, 0.5)} for i in range(10)]

cars += [{'color': random_color(),
          'pos': (-1.5, i * 2.0 - 9.0, 0.0),
          'angle': uniform(-0.5, 0.5)} for i in range(10)]


class ToyCars(Example):
    title = "Toy Cars With Vector Shadow"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_position;
                in vec3 in_normal;

                in vec3 in_color;
                in vec3 in_origin;
                in mat3 in_basis;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec3 v_color;

                void main() {
                    v_vert = in_origin + in_basis * in_position;
                    v_norm = in_basis * in_normal;
                    v_color = in_color;
                    gl_Position = Mvp * vec4(v_vert, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 Light;
                uniform sampler2D Texture;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec3 v_color;

                out vec4 f_color;

                void main() {
                    float lum = clamp(dot(normalize(Light - v_vert), normalize(v_norm)), 0.0, 1.0) * 0.8 + 0.2;
                    f_color = vec4(v_color * lum, 1.0);
                }
            ''',
        )

        self.mvp = self.prog['Mvp']
        self.light = self.prog['Light']

        obj = self.load_scene('lowpoly_toy_car.obj')

        self.vbo = self.ctx.buffer(struct.pack(
            '15f',
            1.0, 1.0, 1.0,
            0.0, 0.0, 0.0,
            1.0, 0.0, 0.0,
            0.0, 1.0, 0.0,
            0.0, 0.0, 1.0,
        ) * len(cars))

        vao_wrapper = obj.root_nodes[0].mesh.vao
        vao_wrapper.buffer(self.vbo, '3f 3f 9f/i', ['in_color', 'in_origin', 'in_basis'])
        self.vao = vao_wrapper.instance(self.prog)

    def render(self, time, frame_time):
        angle = time
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        camera_pos = (np.cos(angle) * 20.0, np.sin(angle) * 20.0, 5.0)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            camera_pos,
            (0.0, 0.0, 0.5),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('f4'))
        self.light.value = camera_pos

        self.vbo.write(b''.join(struct.pack(
            '15f',
            *car['color'],
            *car['pos'],
            1.0, 0.0, 0.0,
            0.0, 1.0, 0.0,
            0.0, 0.0, 1.0,
        ) for car in cars))
        self.vao.render(instances=len(cars))

        self.vbo.write(b''.join(struct.pack(
            '15f',
            0.0, 0.0, 0.0,
            *car['pos'],
            1.0, 0.0, 0.0,
            0.0, 1.0, 0.0,
            0.3, 0.6, 0.0,
        ) for car in cars))
        self.vao.render(instances=len(cars))


if __name__ == '__main__':
    ToyCars.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import GLWindow
# import ModernGL
# import numpy as np

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in float value;

#         void main() {
#             float x = gl_VertexID / 400.0;
#             float y = value / 200.0;

#             gl_Position = vec4(x * 2.0 - 1.0, y, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.3, 0.5, 1.0, 1.0);
#         }
#     '''),
# ])

# data = np.random.geometric(0.05, 400).astype(np.float32)
# vbo = ctx.buffer(data, dynamic=True)

# vao = ctx.simple_vertex_array(prog, vbo, ['value'])

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)

#     data = np.roll(data, -1)
#     vbo.write(data)

#     vao.render(ModernGL.LINE_STRIP)



from PIL import Image, ImageTk


class FramebufferImage(ImageTk.PhotoImage):
    def __init__(self, master, ctx, size):
        super(FramebufferImage, self).__init__(Image.new('RGB', size, (0, 0, 0)))
        self.ctx = ctx
        self.fbo = self.ctx.simple_framebuffer(size)
        self.scope = self.ctx.scope(self.fbo)

    def __enter__(self):
        self.scope.__enter__()

    def __exit__(self, *args):
        self.scope.__exit__(*args)
        self.paste(Image.frombytes('RGB', self.fbo.size, self.fbo.read(), 'raw', 'RGB', 0, -1))



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# import pyglet

# wnd = pyglet.window.Window()
# ctx = ModernGL.create_context()

# vert = ctx.vertex_shader('''
#     #version 330
#     in vec2 vert;
#     void main() {
#         gl_Position = vec4(vert, 0.0, 1.0);
#     }
# ''')

# frag = ctx.fragment_shader('''
#     #version 330
#     out vec4 color;
#     void main() {
#         color = vec4(0.30, 0.50, 1.00, 1.0);
#     }
# ''')

# prog = ctx.program(vert, frag])

# vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])


# @wnd.event
# def on_draw():
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()


# pyglet.app.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# import tkinter as tk

# import ModernGL
# import ModernGL.tk

# ctx = ModernGL.create_standalone_context()

# fbo = ctx.framebuffer(ctx.renderbuffer((512, 512)))

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         uniform float R;
#         uniform float G;
#         uniform float B;

#         out vec4 color;

#         void main() {
#             color = vec4(B, G, R, 1.0);
#         }
#     '''),
# ])

# vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# root = tk.Tk()
# frame = ModernGL.tk.ModernGLCanvas(root, 512, 512)
# frame.pack()


# def redraw(name, value):
#     prog.uniforms[name].value = value

#     fbo.use()
#     fbo.clear(0.94, 0.94, 0.94)
#     vao.render()

#     fbo.read_into(frame.pixels)
#     frame.update()


# R = tk.Scale(root, from_=0, to=255, orient=tk.HORIZONTAL, command=lambda x: redraw('R', int(x) / 255), length=300)
# G = tk.Scale(root, from_=0, to=255, orient=tk.HORIZONTAL, command=lambda x: redraw('G', int(x) / 255), length=300)
# B = tk.Scale(root, from_=0, to=255, orient=tk.HORIZONTAL, command=lambda x: redraw('B', int(x) / 255), length=300)

# R.pack()
# G.pack()
# B.pack()

# R.set(30)
# G.set(80)
# B.set(200)

# root.mainloop()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# from PIL import Image

# # Context

# ctx = ModernGL.create_standalone_context()

# # Framebuffer

# size = (512, 512)
# color_rbo = ctx.renderbuffer(size)
# depth_rbo = ctx.depth_renderbuffer(size)
# fbo = ctx.framebuffer(color_rbo, depth_rbo)

# fbo.use()

# # Shaders & Program

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.3, 0.5, 1.0, 1.0);
#         }
#     '''),
# ])

# # Buffer

# vbo = ctx.buffer(struct.pack(
#     '6f',
#     0.0, 0.8,
#     -0.6, -0.8,
#     0.6, -0.8,
# ))

# # Put everything together

# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# ctx.clear(0.9, 0.9, 0.9)
# vao.render()

# img = Image.frombytes('RGB', size, fbo.read())
# img = img.transpose(Image.FLIP_TOP_BOTTOM)
# img.save('01_hello_world.png')



"""
Example showing the use of PROGRAM_POINT_SIZE
were we modify point sizes int the vertex shader
making us able to specify a unique size per point.

Without PROGRAM_POINT_SIZE enabled the point size
will be static and obtained from ctx.point_size.

We also show the use of gl_PointCoord making us able
to fill the point with any data we want just like a quad.
"""
import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


class Particles(Example):
    title = "Particle System"
    gl_version = (3, 3)
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec3 in_position;

            uniform mat4 projection;
            uniform mat4 modelview;
            uniform float time;

            out vec3 color;

            void main() {
                gl_Position = projection * modelview * vec4(in_position, 1.0);

                // Set the point size
                gl_PointSize = 25 - gl_Position.z + sin((time + gl_VertexID) * 7.0) * 10.0;

                // Calculate a random color based on the vertex index
                color = vec3(mod(gl_VertexID * 432.43, 1.0), mod(gl_VertexID * 6654.32, 1.0), mod(gl_VertexID  * 6544.11, 1.0));
            }
            """,
            fragment_shader="""
            #version 330

            in vec3 color;
            out vec4 outColor;

            void main() {
                // Calculate the distance from the center of the point
                // gl_PointCoord is available when redering points. It's basically an uv coordinate.
                float dist = step(length(gl_PointCoord.xy - vec2(0.5)), 0.5);

                // .. an use to render a circle!
                outColor = vec4(dist * color, dist);
            }
            """,
        )

        positions = np.random.random_sample((1000,)) * 10 - 5
        self.pos_buffer = self.ctx.buffer(positions.astype('f4'))
        self.vao = self.ctx.vertex_array(
            self.prog,
            [(self.pos_buffer, '3f', 'in_position')],
        )
        self.resize(*self.wnd.buffer_size)

    def render(self, time, frame_time):
        self.ctx.enable_only(moderngl.PROGRAM_POINT_SIZE | moderngl.BLEND)
        self.ctx.blend_func = moderngl.ADDITIVE_BLENDING

        rotation = Matrix44.from_eulers((time, time / 2 , time / 3), dtype='f4')
        translation = Matrix44.from_translation((0, 0, -10), dtype='f4')
        modelview = translation * rotation

        self.prog['modelview'].write(modelview)
        self.prog['time'].value = time
        self.vao.render(mode=moderngl.POINTS)

    def resize(self, width, height):
        self.projection = Matrix44.perspective_projection(60, self.wnd.aspect_ratio, 1, 100, dtype='f4')
        self.prog['projection'].write(self.projection)


if __name__ == '__main__':
    Particles.run()



'''
    Renders a rotating and scaling triangle
'''

import numpy as np

from ported._example import Example


class UniformsAndAttributes(Example):
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330
                in vec2 vert;

                uniform vec2 scale;
                uniform float rotation;

                void main() {

                    mat2 rot = mat2(
                        cos(rotation), sin(rotation),
                        -sin(rotation), cos(rotation)
                    );
                    gl_Position = vec4((rot * vert) * scale, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 color;
                void main() {
                    color = vec4(0.3, 0.5, 1.0, 1.0);
                }
            ''',
        )

        self.scale = self.prog['scale']
        self.rotation = self.prog['rotation']

        self.scale.value = (self.wnd.width / self.wnd.height * 0.75, 0.25)

        vertices = np.array([
            1.0, 0.0,
            -0.5, 0.86,
            -0.5, -0.86,
        ], dtype='f4')

        self.vbo = self.ctx.buffer(vertices)
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'vert')

    def render(self, time: float, frame_time: float):
        sin_scale = np.sin(np.deg2rad(time * 60))

        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render()
        self.rotation.value = time

        # Change the scale of the triangle sin-ly
        self.scale.value = (sin_scale * 0.75, 0.75)


if __name__ == '__main__':
    UniformsAndAttributes.run()
